// (C)2007 Attila Kovacs <attila@submm.caltech.edu>


import java.io.*;
import java.util.*;

import nom.tam.fits.*;


public class APEXSZScan extends LabocaScan {
	
	public APEXSZScan() {
	    rcpFile = "aszca/aszca-centred.rcp";
	    pixelDataFile = "aszca/aszca-pixel.dat";
	    channelTemplate = new APEXSZPixel(-1);
	}
	
	public void validate() {
		
		if((!Double.isNaN(minScanSpeed) && minScanSpeed > 0.0) || (!Double.isNaN(maxScanSpeed) && maxScanSpeed > 0.0)) {
			System.err.print("   Clipping frames with unsuitable scanning speeds. ");
			System.err.println(" [" + (int)Math.round(100.0*(float)velocityCut(minScanSpeed, maxScanSpeed)/frames) + "% clipped.]");
		}
		if(!Double.isNaN(maxAcceleration) && maxAcceleration > 0.0) {
			System.err.print("   Clipping frames with high telescope acceleration. ");
			System.err.println(" [" + (int)Math.round(100.0*(float)accelerationCut(maxAcceleration)/frames) + "% clipped.]");
		}	
		aveScanSpeed = getAverageScanningVelocity(0.5 * Unit.s);
		System.err.print("   Median Scanning speed is");
		System.err.println(" " + Util.f1.format(aveScanSpeed/(Unit.arcmin/Unit.s)) + " arcmin/s");
		if(Double.isNaN(zenithTau)) zenithTau = 0.0;
		
		/*
		if(Double.isNaN(zenithTau)) {
			try { zenithTau = tauInterpolator.getValue(0.5*(frame[0].MJD + frame[frames-1].MJD)); } 
			catch(ArrayIndexOutOfBoundsException e) { 
				System.err.println("   WARNING! Outside of tau interpolator range.");	
				zenithTau = 0.0;
			}
		}
		*/
		 
		setZenithTau(zenithTau);	
		System.err.println("   Using zenith tau " + Util.f3.format(zenithTau));

		// Update info on the channels..	
		try { 
		    if(rcpFile == null)  rcpFile = "aszca/aszca-centred.rcp";
		    LabocaPixel.readPixelRCP(channel, rcpFile, channel[referenceChannel]);
		    System.err.println("   Updated pixel position data from " + rcpFile);
		} catch(IOException e) { System.err.println("ERROR! " + e.getMessage()); }
		
		try {   
		    LabocaPixel.loadPixelData(channel, pixelDataFile);
		    System.err.println("   Loaded pixel defaults from " + pixelDataFile);
		} catch(IOException e) { System.err.println(" ERROR! " + e.getMessage()); }
		
		// Channels 320+ are repeats of previous channels.
		// Flag these...
		for(int c=320; c<channels; c++) channel[c].flag |= APEXSZPixel.FLAG_DEAD;

		//APEXSZPixel.markBads(channel);
		
		for(int c=0; c<channels; c++) {
			if(flaggedPixels.contains(channel[c].backendIndex + 1)) channel[c].flag |= APEXSZPixel.FLAG_DEAD;
			channel[c].flag &= APEXSZPixel.FLAG_DEAD;
		}
		
		try { APEXSZPixel.readWiring("aszca" + File.separator + "aszca-wiring.dat"); }
		catch(IOException e) {
			System.err.println("ERROR! Problem reading wiring data. Exiting.");
			System.exit(1);
		}
		
		
		APEXSZPixel.fillWiringInfo(channel);
		System.err.println("   Loaded pixel wiring data.");

		trim();
		slim();
		
		allChannels = APEXSZPixel.getAllChannels(channel);

		try { 
			boxChannels = APEXSZPixel.getChannelGroups(channel, APEXSZPixel.class.getField("box")); 
			cableChannels = APEXSZPixel.getChannelGroups(channel, APEXSZPixel.class.getField("cable"));
			ampChannels = APEXSZPixel.getChannelGroups(channel, APEXSZPixel.class.getField("amp")); 
			waferChannels = APEXSZPixel.getChannelGroups(channel, APEXSZPixel.class.getField("wafer"));
			regionChannels = APEXSZPixel.getChannelGroups(channel, APEXSZPixel.class.getField("region"));
		}
		catch(Exception e) { e.printStackTrace(); }
	
		pinOrder();
		yOrder();
		
		APEXSZPixel.findNeighbours(channel, 2.0*Unit.arcmin, 5.0*Unit.arcmin);
		//for(int c=0; c<channels; c++) ((APEXSZPixel) channel[c]).wedgeOnlyNeighbours();
		System.err.println("   Registered pixel neighbours.");

		if(!Double.isNaN(hipassTimeScale)) {
			System.err.print("   Hi-pass Filtering...");
			highPassFilter(hipassTimeScale);
			System.err.println();
		}

		if(isChopped) markChopped();
	}
	
	
	public LabocaScan[] readScanDirectory(LabocaScan common, String dir, String ext) throws IOException, FitsException, HeaderCardException {
		int subscans = common.readScanInfo(dir + File.separator + "SCAN" + ext);
		common.readPar(dir + File.separator + "BOLOSZ-SZACBE-FEBEPAR" + ext);
		
		Vector<APEXSZScan> buffer = new Vector<APEXSZScan>();

		for(int i=0; i<subscans; i++) {
			try {
				APEXSZScan subscan = (APEXSZScan) common.clone();

				System.err.println(" Subscan {" + (i+1) + "}:");
				subscan.subscanNo = i;
				subscan.readDataPar(dir + File.separator + (i+1) + File.separator + "BOLOSZ-SZACBE-DATAPAR" + ext);
				subscan.readData(dir + File.separator + (i+1) + File.separator + "BOLOSZ-SZACBE-ARRAYDATA-1" + ext);
				subscan.readMonitor(dir + File.separator + (i+1) + File.separator + "MONITOR" + ext);
				subscan.validate();

				buffer.add(subscan);
			}
			catch(Exception e) {
				System.err.println(" ERROR: " + e.getMessage());
				e.printStackTrace();
			}
		}

		APEXSZScan[] set = new APEXSZScan[buffer.size()];
		for(int i=0; i<buffer.size(); i++) set[i] = buffer.get(i);
		subscans = buffer.size();

		return set;
	}

	
	public void yOrder() {
		for(int g=0; g<regionChannels.length; g++) {
			Vector<LabocaPixel> selection = new Vector<LabocaPixel>();
			for(int k=0; k<regionChannels[g].length; k++) selection.add(channel[regionChannels[g][k]]);
			
			Comparator<LabocaPixel> pinorder = new Comparator<LabocaPixel>() {
				public int compare(LabocaPixel c1, LabocaPixel c2) {
					return Double.compare(c1.position.y, c2.position.y);
				}				
			};
			
			Collections.sort(selection, pinorder);
			
			for(int k=0; k<regionChannels[g].length; k++) regionChannels[g][k] = selection.get(k).index;		
		}
		System.err.println("   Regions rearranged to y-order.");	
	}
	

	/*
	public void readMonitor(String fileName)  throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));

		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
		readMonitor(hdu);
	}
	*/
	
	
	public void readMonitor(String fileName)  throws IOException, FitsException, HeaderCardException {
		
	}
	
	public void readMonitor(BinaryTableHDU hdu)  throws IOException, FitsException, HeaderCardException {
	}
	
	
	/*
	public void readMonitor(BinaryTableHDU hdu)  throws IOException, FitsException, HeaderCardException {
		System.err.print("   Reading Monitor Table...");
		Header header = hdu.getHeader();
		hdu.getData().getData();
		
		int iMJD = hdu.findColumn("MJD");
		int iLABEL = hdu.findColumn("MONPOINT");
		int iVALUE = hdu.findColumn("MONVALUE");

		int n = header.getIntValue("NAXIS2");
		for(int i=0; i<n; i++) {
			Object[] row = hdu.getRow(i);
			if(((String) row[iLABEL]).endsWith("USTAGETEMP")) {
				System.err.println(row[iLABEL] + "\t" + ((double[]) row[iVALUE])[0]);
			}
		}

		
	}
	*/
	
}
