// (C)2007 Attila Kovacs <attila@submm.caltech.edu>

import java.io.*;
import java.util.*;


import nom.tam.fits.*;
import nom.tam.util.*;


public class LabocaScan implements Cloneable {
	LabocaFrame[] frame;
	LabocaPixel[] channel;
	int channels, frames;	

	LabocaPixel channelTemplate = new LabocaPixel(-1);
	
	EquatorialCoordinates equatorial;
	HorizontalCoordinates horizontal;
	
	boolean isNativeEquatorial = false;
	
	double PA;
	String sourceName;
	String scanPath;
	String descriptor;
	String comments = new String();
	int serialNo, subscanNo;
	
	double beamFWHM = 9.0*Unit.arcsec;
	
	boolean debug = false;
	double G; // conversion from counts to V at bol stage.
	double janskyPerMicroV;
	
	SkyMap map;

	boolean approximateSourceMap = false, extendedFilter = false;
	int sourceGeneration = 0;
	double extendedScale = 120.0 * Unit.arcsec;

	double minScanSpeed = Double.NaN;
	double maxScanSpeed = Double.NaN;
	double maxAcceleration = Double.NaN;
	double aveScanSpeed = Double.NaN;
	
	int referenceChannel;

	double integrationTime = 1.0;

	int onPhase = 1;
	int offPhase = 2;

	boolean isChopped = false;
	double chopperThrow = 0.0;

	double zenithTau = Double.NaN;
	double nefd = 0.0; // It is readily cast into the Jy sqrt(s) units!!!
	double weight = 1.0;
	
	int[] allChannels;
	int[][] boxChannels, cableChannels, ampChannels, waferChannels, regionChannels;
	
	boolean useHe3Data, calcHe3Gains;		
	double he3TimeScale = 3.0 * Unit.sec/Unit.day;
	
	boolean useBlindData, calcBlindGains;		
	double blindTimeScale = 3.0 * Unit.sec/Unit.day;
	
	double hipassTimeScale = Double.NaN;
	boolean scramble = false;
	int driftN = 128; // The 1/f filtering timescale (in frames), or its default value to be assumed...
	
	int goodChannelCount = -1;
	
	public String rcpFile;
	public String pixelDataFile = "laboca/laboca-pixel.dat";
	
	public Vector<Integer> flaggedPixels = new Vector<Integer>();
	
	public LabocaScan() {}
	
	public Object clone() { 
		try { 
			LabocaScan clone = (LabocaScan) super.clone(); 
			clone.mapIndex = new short[2];
			clone.reuseOffset = new Vector2D();
			clone.offsetData = null;
			clone.corrData = null;
			clone.gainData = null;
			return clone;
		}	
		catch(CloneNotSupportedException e) { return null; }	
	}
	
	public LabocaScan copy() {
		LabocaScan copy = (LabocaScan) clone();
		copy.channel = new LabocaPixel[channel.length];
		for(int c=0; c<channel.length; c++) copy.channel[c] = (LabocaPixel) channel[c].clone();
		copy.comments = new String(comments);
		return copy;
	}
	
	public void validate() {	
		// Finish with updating info on the channels...
		
		if((!Double.isNaN(minScanSpeed) && minScanSpeed > 0.0) || (!Double.isNaN(maxScanSpeed) && maxScanSpeed > 0.0)) {
			System.err.print("   Clipping frames with unsuitable scanning speeds. ");
			System.err.println(" [" + (int)Math.round(100.0*(float)velocityCut(minScanSpeed, maxScanSpeed)/frames) + "% clipped.]");
		}
		if(!Double.isNaN(maxAcceleration) && maxAcceleration > 0.0) {
			System.err.print("   Clipping frames with high telescope acceleration. ");
			System.err.println(" [" + (int)Math.round(100.0*(float)accelerationCut(maxAcceleration)/frames) + "% clipped.]");
		}	
		aveScanSpeed = getAverageScanningVelocity(0.5 * Unit.s);
		System.err.print("   Median Scanning speed is");
		System.err.println(" " + Util.f1.format(aveScanSpeed/(Unit.arcmin/Unit.s)) + " arcmin/s");
		
		if(Double.isNaN(zenithTau)) {
			try { zenithTau = tauInterpolator.getValue(0.5*(frame[0].MJD + frame[frames-1].MJD)); } 
			catch(ArrayIndexOutOfBoundsException e) { 
				System.err.println("   WARNING! Outside of tau interpolator range.");	
				zenithTau = 0.0;
			}
		}
		setZenithTau(zenithTau);	
		System.err.println("   Using zenith tau " + Util.f3.format(zenithTau));
		
		try { 
			if(rcpFile == null) {
				if(frame[0].MJD < 54522) rcpFile = "laboca/master-laboca-oct07.rcp";
				else rcpFile = "laboca/master-laboca-feb08.rcp";
			}
			LabocaPixel.readPixelRCP(channel, rcpFile, channel[referenceChannel]);
			System.err.println("   Updated pixel position data from " + rcpFile);
		} catch(IOException e) { System.err.println("ERROR! " + e.getMessage()); }
		
		try {
			LabocaPixel.loadPixelData(channel, pixelDataFile);
			System.err.println("   Loaded pixel defaults from " + pixelDataFile);
		} catch(IOException e) { System.err.println(" ERROR! " + e.getMessage()); }
		
		for(int c=0; c<channels; c++) {
			if(flaggedPixels.contains(channel[c].backendIndex + 1)) channel[c].flag |= LabocaPixel.FLAG_DEAD;
			channel[c].flag &= LabocaPixel.FLAG_DEAD;
		}
		
		try { LabocaPixel.readWiring("laboca" + File.separator + "laboca-wiring.dat"); }
		catch(IOException e) {
			System.err.println("ERROR! Problem reading wiring data. Exiting.");
			System.exit(1);
		}
		
		LabocaPixel.fillWiringInfo(channel);
		System.err.println("   Loaded pixel wiring data.");
		
		
		if(calcHe3Gains) {
			try { getHe3TemperatureGains(); } 
			catch(IOException e) { }	
		}
		
		if(useHe3Data) {
			try { correctForTemperatureDrifts(); } 
			catch(IOException e) {}	
		}
		
		else if(useBlindData) {
			try { correctViaBlinds(); } 
			catch(IOException e) {}	
		}
		
		LabocaPixel.markBads(channel);
		
		trim();
		slim();
	
		allChannels = LabocaPixel.getAllChannels(channel);
		try { 
			boxChannels = LabocaPixel.getChannelGroups(channel, LabocaPixel.class.getField("box")); 
			cableChannels = LabocaPixel.getChannelGroups(channel, LabocaPixel.class.getField("cable")); 
			ampChannels = LabocaPixel.getChannelGroups(channel, LabocaPixel.class.getField("amp")); 
		}
		catch(Exception e) { e.printStackTrace(); }
		
		pinOrder();
		
		LabocaPixel.findNeighbours(channel, 24*Unit.arcsec, 48.0*Unit.arcsec);
		System.err.println("   Registered pixel neighbours.");
		
		if(!Double.isNaN(hipassTimeScale)) {
			System.err.print("   Hi-pass Filtering...");
			highPassFilter(hipassTimeScale);
			System.err.println();
		}
		
		if(isChopped) markChopped();
		
		
	}
	
	public void pinOrder() {
		for(int g=0; g<cableChannels.length; g++) {
			Vector<LabocaPixel> selection = new Vector<LabocaPixel>();
			for(int k=0; k<cableChannels[g].length; k++) selection.add(channel[cableChannels[g][k]]);
			
			Comparator<LabocaPixel> pinorder = new Comparator<LabocaPixel>() {
				public int compare(LabocaPixel c1, LabocaPixel c2) {
					if(c1.pin < c2.pin) return -1;
					else if(c1.pin > c2.pin) return 1;
					else return 0;
				}				
			};
			
			Collections.sort(selection, pinorder);
			
			for(int k=0; k<cableChannels[g].length; k++) cableChannels[g][k] = selection.get(k).index;		
		}
		System.err.println("   Cables rearranged to pin order.");	
	}
	
	public void setZenithTau(double value) {
		zenithTau = value;
		for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].setZenithTau(zenithTau);
	}
	
	public void trim() {
		int first = 0, last = frames-1;
		while(frame[first] == null && first < frames-1) first++;
		while(frame[last] == null && last > first) last--;
	
		if(first == 0 && last == frames-1) return;
		
		LabocaFrame[] trimmed = new LabocaFrame[last-first+1];
		for(int from=first, to=0; from <= last; from++, to++) trimmed[to] = frame[from];
		
		frame = trimmed;
		frames = trimmed.length;	
		
		System.err.println("   Trimmed to " + frames + " frames.");
	}
	
	public void slim() {
		int good = 0;
		for(int c=0; c<channels; c++) if(channel[c].flag == 0 || c == referenceChannel) good++;
		
		LabocaPixel[] goodChannel = new LabocaPixel[good];
		
		for(int c=0, k=0; c<channels; c++) if(channel[c].flag == 0 || c == referenceChannel) goodChannel[k++] = channel[c];		
		
		//for(int k=0; k<goodChannel.length; k++) System.err.println(goodChannel[k].index + " ---> " + k);
		
		for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].slim(goodChannel);
		
		channel = goodChannel;		
		channels = channel.length;	
		
		for(int c=0; c<channels; c++) {
			if(channel[c].backendIndex == referenceChannel) referenceChannel = c;
			channel[c].index = c;
		}
	
		
		System.err.println("   Slimmed to " + channels + " working pixels.");
	}
	
	
	public void scale(double factor) {
		for(int t=0; t<frames; t++) for(int c=0; c<channels; c++) frame[t].data[c] *= factor;
	}
	
	public int[][] getChannelGroup(String name) {
		if(name.equalsIgnoreCase("all")) return new int[][] { allChannels };
		else if(name.equalsIgnoreCase("boxes")) return boxChannels;
		else if(name.equalsIgnoreCase("cables")) return cableChannels;
		else if(name.equalsIgnoreCase("amps")) return ampChannels;
		else if(name.equalsIgnoreCase("wafers")) return waferChannels;
		else if(name.equalsIgnoreCase("regions")) return regionChannels;
		else return null;
	}

	// Recycle the index array
	private short[] mapIndex = new short[2];
	public final short[] getMapIndexes(int t, int bol) {
		Vector2D coord = getMapOffsets(t, bol);

		mapIndex[0] = (short)map.indexOfdX(coord.x);
		mapIndex[1] = (short)map.indexOfdY(coord.y);

		return mapIndex;	
	}

	private Vector2D reuseOffset = new Vector2D();
	public final Vector2D getMapOffsets(int t, int bol) {
		Vector2D mapOffset = reuseOffset;

		if(!map.isHorizontal) map.projection.project(frame[t].getCoords(channel[bol]), mapOffset);
		else mapOffset = frame[t].getdAZdEL(channel[bol]);

		return mapOffset;
	}	


	public void correctForTemperatureDrifts() throws IOException {
		System.err.println("   Correcting for temperature drifts.");

		double sum = 0.0;
		int n=0;
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			sum += frame[t].he3Temp;
			n++;
		}
		double aveHe3Temp = sum/n;

		// Read gains into LabocaPixel -> temperatureGain:
		BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream("laboca" + File.separator + "laboca-he3-gains.dat")));
		String line;
		while((line = in.readLine()) != null) if(line.length() > 0) if(line.charAt(0) != '#') {
			StringTokenizer tokens = new StringTokenizer(line);
			int c = Integer.parseInt(tokens.nextToken()) - 1;
			channel[c].temperatureGain = Double.parseDouble(tokens.nextToken());
		}

		// Apply correction
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			double dT = frame[t].he3Temp - aveHe3Temp;
			float[] readout = frame[t].data;
			for(int c=0; c<channels; c++) readout[c] -= channel[c].temperatureGain * dT;
		}
	}
	
	public void getHe3TemperatureGains() throws IOException {
		System.err.println("   Calculating He3 Temperature Gains.");
		
		double sum = 0.0;
		int n=0;
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			sum += frame[t].he3Temp;
			n++;
		}
		double aveHe3Temp = sum/n;
	
		// Level the signals...
		removeOffsets(LabocaPixel.getAllChannels(channel), 0, frames);
		
		for(int c=0; c<channels; c++) {
			double sumIT = 0.0, sumT2 = 0.0;
			for(int t=0; t<frames; t++) if(frame[t] != null) {
				double T = frame[t].he3Temp - aveHe3Temp;
				sumIT += frame[t].data[c] * T;
				sumT2 += T * T;
			}
			channel[c].temperatureGain = sumIT / sumT2;
		}
		
		// Now write to a file
		String fileName = MiniCRUSH.workPath + File.separator + "he3-gains-" + serialNo + ".dat";
		
		PrintWriter out = new PrintWriter(new BufferedOutputStream(new FileOutputStream(fileName)));
		out.println("# He3 Temperature Gains Data.");
		out.println("# Measured by MiniCrush ver " + MiniCRUSH.version);
		out.println("# ");
		out.println("# BEch\tGain");
		out.println("#     \t(V/K)");
		out.println("# ----\t-----");
		for(int c=0; c<channels; c++) out.println((channel[c].backendIndex + 1) + "\t" + Util.e6.format(channel[c].temperatureGain));
		
		out.flush();
		out.close();
		System.err.println(" Written gain data to " + fileName + ".");
		System.err.println(" Exiting.");
		
		System.exit(0);
	}

	
	public void correctViaBlinds() throws IOException {
		System.err.println("   Correcting for temperature drifts (Blind bolometers).");

		int[] blindChannels = { 3, 162 };
		float[] gain = { (float) channel[3].temperatureGain, (float) channel[162].temperatureGain };
		float[] signal = new float[frames];
		float[] smoothed = new float[frames];
		
		robustRemoveOffsets(blindChannels, 0, frames);
		getCorrelated(blindChannels, gain, signal);
		
		int smoothT = Math.min((int)Math.ceil(blindTimeScale / integrationTime), frames);
		double sum = 0.0;
		for(int t=0; t<smoothT-1; t++) sum += signal[t];
		
		for(int fromt=smoothT, tot=smoothT/2; fromt<smoothT; fromt++, tot++) {
			sum += signal[fromt];
			smoothed[tot] = (float) sum / smoothT;
			sum -= signal[fromt-smoothT];
		}
		for(int t=0; t<smoothT/2; t++) smoothed[t] = smoothed[smoothT/2];
		for(int t=frames-smoothT/2; t<frames; t++) smoothed[t] = smoothed[frames-smoothT/2];
		
			
		// Read gains into LabocaPixel -> temperatureGain:
		BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream("he3-gains.dat")));
		String line;
		while((line = in.readLine()) != null) if(line.length() > 0) if(line.charAt(0) != '#') {
			StringTokenizer tokens = new StringTokenizer(line);
			int c = Integer.parseInt(tokens.nextToken()) - 1;
			channel[c].temperatureGain = Double.parseDouble(tokens.nextToken());
		}

		// Apply correction
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			float[] readout = frame[t].data;
			for(int c=0; c<channels; c++) readout[c] -= channel[c].temperatureGain * smoothed[t];
		}
	}
	
	
	
	public void removeOffsets(int[] channelSelection, int fromt, int tot, boolean robust) {
		if(robust) robustRemoveOffsets(channelSelection, fromt, tot);
		else removeOffsets(channelSelection, fromt, tot);
	}
		
	private void removeOffsets(final int[] channelSelection, final int fromt, int tot) {
		tot = Math.min(tot, frames);

		//checkForNaNs();
		
		for(int k=0; k<channelSelection.length; k++) {
			final LabocaPixel pixel = channel[channelSelection[k]];

			final int c = pixel.index;
			double offset = 0.0; 
			int n=0;

			for(int t=fromt; t<tot; t++) {
				final LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0) {
					offset += exposure.data[pixel.index];
					n++;
				}
			}

			if(n > 0) {
				offset /= n;
				for(int t=fromt; t<tot; t++) if(frame[t] != null) frame[t].data[c] -= offset;
				pixel.dependents++;
			}
		}
		
		if(debug) checkForNaNs();
	}

	
	
	private float[] offsetData;

	private void robustRemoveOffsets(final int[] channelSelection, final int fromt, int tot) {
		tot = Math.min(tot, frames);

		if(offsetData == null) offsetData = new float[tot-fromt];
		if(offsetData.length < tot-fromt) offsetData = new float[tot-fromt];
		
		for(int k=0; k<channelSelection.length; k++) {
			final LabocaPixel pixel = channel[channelSelection[k]];

			final int c = pixel.index;
			int n = 0;

			for(int t=fromt; t<tot; t++) {
				final LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0)
					offsetData[n++] = exposure.data[c];
			}

			if(n > 0) {
				final double offset = ArrayUtil.median(offsetData, 0, n);
				for(int t=fromt; t<tot; t++) if(frame[t] != null) frame[t].data[c] -= offset;
				pixel.dependents++;
			}
		}
		if(debug) checkForNaNs();
	}

	public void removeNeighbours(boolean useGains) {
		comments += "N";
		
		// First find the total neighbour weight norms...
		double[] sumw = new double[channels];
		double[] w = new double[channels];
		double[] g = new double[channels];
		
		for(int c=0; c<channels; c++) {
			w[c] = channel[c].weight;
			g[c] = useGains ? channel[c].gain : 1.0;
		}
		
		for(int c1=0; c1<channels; c1++) if(!channel[c1].isBad) for(LabocaPixel ch2 : channel[c1].neighbours) if(!ch2.isBad) {
			if(ch2.flag == 0) sumw[c1] += w[ch2.index] * g[ch2.index] * g[ch2.index];
			if(channel[c1].flag == 0) sumw[ch2.index] += w[c1] * g[c1] * g[c1];
		}
		
		float[] data = new float[channels];
		for(int t=0; t<frames; t++) {
			LabocaFrame exposure = frame[t];
			if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) {
				float[] readout = frame[t].data;
				for(int c=0; c<channels; c++) data[c] = readout[c]; 

				for(int c1=0; c1<channels; c1++) if(!channel[c1].isBad) if(exposure.sampleFlag[c1] == 0) {
					for(LabocaPixel ch2 : channel[c1].neighbours) if(!ch2.isBad) if(exposure.sampleFlag[ch2.index] == 0) {
						if(channel[c1].neighbours.size() > 1 && ch2.flag == 0 && sumw[c1] > 0.0) 
							readout[c1] -= w[ch2.index] * g[ch2.index] * data[ch2.index] / sumw[c1];
						if(ch2.neighbours.size() > 1 && channel[c1].flag == 0 && sumw[ch2.index] > 0.0)
							readout[ch2.index] -= w[c1] * g[c1] * data[c1] / sumw[ch2.index];
					}
				}
			}
		}
		
	}
	
	public void getCorrelated(int[] channelSelection, float[] gain, float[] correlated, boolean robust) {
		int n = 0;
		for(int k=0; k<channelSelection.length; k++) {
			LabocaPixel pixel = channel[channelSelection[k]];
			if(pixel.flag == 0 && pixel.weight > 0.0) n++; 
		
		}
		
		if(n <= 1) return;
		
		if(robust) getRobustCorrelated(channelSelection, gain, correlated);
		else {
			getCorrelated(channelSelection, gain, correlated);
			calcGroupFiltering(channelSelection, gain);
		}
	}
		
	// Get correlated for all frames even those that are no good...
	// But use only channels that are valid, and skip over flagged samples...
	private void getCorrelated(int[] channelSelection, float[] gain, float[] correlated) {
		if(channelSelection.length < 2) return;
		
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			final LabocaFrame exposure = frame[t];
			double sum = 0.0, sumwG2 = 0.0;
			
			for(int k=0; k<channelSelection.length; k++) {
				final LabocaPixel pixel = channel[channelSelection[k]];
				final float G = gain[k];
				
				if(pixel.flag == 0) if(exposure.sampleFlag[pixel.index] == 0) {
					sum += pixel.weight * G * exposure.data[pixel.index];
					sumwG2 += pixel.weight * G * G;
				}
			}
			if(sumwG2 > 0.0) {
				final double dC = sum / sumwG2;
				correlated[t] += dC;
				
				for(int k=0; k<channelSelection.length; k++) {
					final LabocaPixel pixel = channel[channelSelection[k]];
					final float G = gain[k];
					exposure.data[pixel.index] -= G * dC;
					if(pixel.flag == 0) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) 
						if(exposure.sampleFlag[pixel.index] == 0) pixel.dependents += pixel.weight * G * G / sumwG2;
					
				}
			}
		}
		if(debug) checkForNaNs();
	}
	
	private WeightedPoint[] corrData;
	private void getRobustCorrelated(final int[] channelSelection, final float[] gain, final float[] correlated) {
		if(channelSelection.length < 2) return;
		
		if(corrData == null) corrData = new WeightedPoint[channelSelection.length];
		else if(corrData.length < channelSelection.length) corrData = new WeightedPoint[channelSelection.length];
		if(corrData[0] == null) for(int i=0; i<corrData.length; i++) corrData[i] = new WeightedPoint();
		
		for(int t=0; t<frames; t++) if(frame[t] != null) {
			final LabocaFrame exposure = frame[t];
			int n=0;
			double sumw = 0.0;
			
			for(int k=0; k<channelSelection.length; k++) {
				final LabocaPixel pixel = channel[channelSelection[k]];
				final float G = gain[k];
				
				if(pixel.flag == 0) if(exposure.sampleFlag[pixel.index] == 0) {
					final double weight = pixel.weight * G * G;
					corrData[n].value = exposure.data[pixel.index] / G;
					corrData[n++].weight = weight;
					sumw += weight;
				}
			}
			if(n > 0) {
				final double dC = ArrayUtil.median(corrData, 0, n);
				correlated[t] += dC;
				
				for(int k=0; k<channelSelection.length; k++) {
					final LabocaPixel pixel = channel[channelSelection[k]];
					final float G = gain[k];
					exposure.data[pixel.index] -= G * dC;
					
					if(pixel.flag == 0) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) 
						if(exposure.sampleFlag[pixel.index] == 0) pixel.dependents += pixel.weight * G * G / sumw;
					
				}
			}
		}
		if(debug) checkForNaNs();
	}
	
	
	public void getGains(int[] channelSelection, float[] gain, 
			float[] correlated, double loLimit, double hiLimit, int flagPattern, boolean sourceSync, int normMethod, boolean robust) {	
		
		int n = 0;
		for(int k=0; k<channelSelection.length; k++) {
			LabocaPixel pixel = channel[channelSelection[k]];
			if(pixel.flag == 0 && pixel.weight > 0.0) n++; 
		
		}
		
		if(n <= 1) return;
		
		if(robust) getRobustGains(channelSelection, gain, correlated, loLimit, hiLimit, flagPattern, sourceSync, normMethod);
		else getGains(channelSelection, gain, correlated, loLimit, hiLimit, flagPattern, sourceSync, normMethod);
	}
	

	
	// Calculate gains for *ALL* channels, even if bad.
	private void getGains(final int[] channelSelection, final float[] gain, 
		final float[] correlated, final double loLimit, final double hiLimit, final int flagPattern, final boolean sourceSync, final int normMethod) {
		if(channelSelection.length < 2) return;

		// override the normalization method...
		//normMethod = 0;
		
		final double[] dG = new double[gain.length];

		for(int k=0; k<channelSelection.length; k++) {
			final LabocaPixel pixel = channel[channelSelection[k]];
			double RC = 0.0, C2 = 0.0;

			for(int t=0; t<frames; t++) {
				final LabocaFrame exposure = frame[t];
				final double C = correlated[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[pixel.index] == 0) {
					RC += exposure.data[pixel.index] * C;
					C2 += C * C;
				}
			}
			if(C2 > 0.0) dG[k] = RC / C2;
		}

		// Important!!! Must synch with the maximum likelihood estimates
		// If renormalizing, then correlated has to be derived again before can be injected...
		syncGains(channelSelection, gain, dG, correlated, sourceSync);	
		normGainChanges(channelSelection, gain, dG, hiLimit, loLimit, normMethod);
		flagGains(channelSelection, gain, loLimit, hiLimit, flagPattern, normMethod); 	
		//syncGains(channelSelection, gain, dG, correlated, sourceSync);			
	}
	
	
	private WeightedPoint[] gainData;
	private void getRobustGains(final int[] channelSelection, final float[] gain, 
			final float[] correlated, final double loLimit, final double hiLimit, final int flagPattern, final boolean sourceSync, final int normMethod) {
		if(channelSelection.length < 2) return;

		// override the normalization method...
		//normMethod = 0;

		final double[] dG = new double[gain.length];
		
		if(gainData == null) gainData = new WeightedPoint[frames];
		if(gainData.length < frames) gainData = new WeightedPoint[frames];
		
		if(gainData[0] == null) for(int t=0; t<frames; t++) gainData[t] = new WeightedPoint();

		for(int k=0; k<channelSelection.length; k++) {
			final LabocaPixel pixel = channel[channelSelection[k]];
			int n=0;

			for(int t=0; t<frames; t++) {
				final LabocaFrame exposure = frame[t];
				final double C = correlated[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[pixel.index] == 0) {
					gainData[n].value = exposure.data[pixel.index] / C;
					if(!Double.isNaN(gainData[n].value)) gainData[n++].weight = C * C;
				}
			}
			dG[k] = n > 0 ? ArrayUtil.median(gainData, 0, n) : 0.0;
		}

		// Important!!! Must synch with the maximum likelihood estimates
		// If renormalizing, then correlated has to be derived again before it can be injected...
		syncGains(channelSelection, gain, dG, correlated, sourceSync);
		normGainChanges(channelSelection, gain, dG, hiLimit, loLimit, normMethod);
		flagGains(channelSelection, gain, loLimit, hiLimit, flagPattern, normMethod); 
		//syncGains(channelSelection, gain, dG, correlated, sourceSync);

	}
	
	

	
	
		
	private void normGainChanges(int[] channelSelection, float[] gain, double[] dG, double hiLimit, double loLimit, int normMethod) {
		// Normalize gain change by the average of pixels that see light and have not already been flagged for abnormal gains...
		
		int pixelFlags = LabocaPixel.FLAG_DEAD | LabocaPixel.FLAG_BLIND | LabocaPixel.FLAG_GAIN;
		
		WeightedPoint[] point = new WeightedPoint[channelSelection.length];
		for(int i=0; i<channelSelection.length; i++) point[i] = new WeightedPoint();
		int n=0;
		
		if(normMethod == LabocaPixel.GAINS_POSITIVE) {
			for(int k=0; k<channelSelection.length; k++) {
				LabocaPixel pixel = channel[channelSelection[k]];
				if((pixel.flag & pixelFlags) == 0) {
					point[n].value = dG[k] / gain[k];
					point[n++].weight = pixel.weight * gain[k] * gain[k];
				}
			}
			if(n > 0) {
				double avedG = ArrayUtil.median(point, 0, n);
				for(int k=0; k<channelSelection.length; k++) dG[k] -= avedG;
			}
		}
		else if(normMethod == LabocaPixel.GAINS_BIDIRECTIONAL) {		
			for(int k=0; k<channelSelection.length; k++) {
				LabocaPixel pixel = channel[channelSelection[k]];
				if((pixel.flag & pixelFlags) == 0) {
					point[n].value = Math.abs(gain[k] + dG[k]); 
					point[n++].weight = pixel.weight * gain[k] * gain[k];
				}
			}
			if(n > 0) {
				double norm = ArrayUtil.median(point, 0, n);
				for(int k=0; k<channelSelection.length; k++) dG[k] = (gain[k] + dG[k]) / norm - gain[k];
			}
		}
		
		// Report if a gain change is unrealistic...
		/*
		for(int k=0; k<channelSelection.length; k++) {
			LabocaPixel pixel = channel[channelSelection[k]];
			if((pixel.flag & pixelFlags) == 0) if(Math.abs(dG[k]) > 1.0) System.err.println(" WARNING! dG[" + (pixel.index+1) + "]: " + Util.f1.format(dG[k]));
		}
		*/
	}
	
	private void syncGains(int[] channelSelection, float[] gain, double[] dG, float[] correlated, boolean sourceSync) {
		// Sync to all channels in selection.
		for(int k=0; k<channelSelection.length; k++) {
			LabocaPixel pixel = channel[channelSelection[k]];
			gain[k] += dG[k];	    
			
			// Sync to all frames
			for(int t=0; t<frames; t++) if(frame[t] != null) {
				LabocaFrame exposure = frame[t];
				exposure.data[pixel.index] -= dG[k] * correlated[t];

				if(sourceSync && sourceGeneration > 0) if((exposure.flag & LabocaFrame.SKIP_SOURCE) == 0) {
					final double frameGain = exposure.transmission; 
					final short[] index = getMapIndexes(t, pixel.index);
					exposure.data[pixel.index] -= dG[k] * frameGain * map.signal[index[0]][index[1]];
				}
			}
			pixel.dependents++;
		}
	}

	private void flagGains(int[] channelSelection, float[] gain, double loLimit, double hiLimit, int flagPattern, int normMethod) {
		// Flag/unFlag
		if(flagPattern > 0) for(int k=0; k<channelSelection.length; k++) {
			LabocaPixel pixel = channel[channelSelection[k]];
			
			double G = Double.NaN;
			if(normMethod == LabocaPixel.GAINS_POSITIVE) G = gain[k];
			else if(normMethod == LabocaPixel.GAINS_BIDIRECTIONAL) G = Math.abs(gain[k]);

			if(G < loLimit || G > hiLimit) pixel.flag |= flagPattern;
			else pixel.flag &= ~flagPattern;
		}
		
		if(debug) checkForNaNs();
	}

	
	
	private void applyGains(int[] channelSelection, float[] channelGain) {
		for(int k=0; k<channelSelection.length; k++) channel[channelSelection[k]].gain = channelGain[k];
	}


	public void census() {
		goodChannelCount = 0;
		for(int c=0; c<channels; c++) if(!channel[c].isBad) if(channel[c].flag == 0 && channel[c].weight > 0.0) goodChannelCount++;
		if(debug) checkForNaNs();

	}

	public void getPixelWeights() {
		comments += "W";

		//checkForNaNs();

		for(int c=0; c<channels; c++) if(!channel[c].isBad) {
			int points = 0;
			double var=0.0;
			
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.WEIGHTING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0) {
					double value = exposure.data[c];
					var += value*value;
					points++;
				}
			}
			
			channel[c].dof = points - channel[c].dependents;
			channel[c].weight = channel[c].dof / var;
		}

		flagWeights(0.03, 30.0);
		
		census();
	}

	public void getDifferencialPixelWeights() {
		comments += "w";
		int delta = 10;

		//checkForNaNs();

		for(int c=0; c<channels; c++) if(!channel[c].isBad) {
			double var=0.0;
			int points = 0;

			for(int t=delta; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				LabocaFrame prior = frame[t-delta];
				if(exposure != null && prior != null) if(((exposure.flag | prior.flag) & LabocaFrame.WEIGHTING_FLAGS) == 0)
					if(exposure.sampleFlag[c] == 0) if(prior.sampleFlag[c] == 0) {
						double value = exposure.data[c] - prior.data[c];
						var += value*value;
						points++;
					}		    
			}

			channel[c].dof = points - channel[c].dependents;
			channel[c].weight = 2.0 * channel[c].dof / var;			
		}
 
		flagWeights(0.03, 30.0);
		census();
		
	}

	
	public void getRobustPixelWeights() {
		comments += "[W]";
	
		double[] dev2 = new double[frames];

		for(int c=0; c<channels; c++) if(!channel[c].isBad) {
			int points = 0;
			
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.WEIGHTING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0) {
					double value = frame[t].data[c];
					dev2[points++] = value*value;
				}		    
			}	
				
			channel[c].dof = points - channel[c].dependents;
			channel[c].weight = channel[c].dof / ((float)points * ArrayUtil.median(dev2, 0, points) / 0.454937);		
		}
		
		flagWeights(0.03, 30.0);
		
		census();
	}
	
	public void noDependents() {
		for(int c=0; c<channels; c++) if(!channel[c].isBad) channel[c].dependents = 0;
	}
	
	// Changed to flag with gains!!!
	public void flagWeights(double min, double max) {
		if(goodChannelCount == 0) {
			comments += "(----)";
			return;			
		}
		
		// Flag channels with insufficient degrees of freedom
		double[] weights = new double[channels];
		int n=0;
		for(int c=0; c<channels; c++) if(!channel[c].isBad) {

			if(channel[c].dof > 0.0) {
				channel[c].flag &= ~LabocaPixel.FLAG_DOF;
				if((channel[c].flag & ~LabocaPixel.FLAG_SENSITIVITY) == 0)
					weights[n++] = channel[c].weight * channel[c].gain * channel[c].gain;
			}
			else {
				channel[c].flag |= LabocaPixel.FLAG_DOF;
				channel[c].weight = 0.0;
			}
			
		}
				
		// Flag out channels with unrealistically small or large weights
		
		double medianWeight = n > 0 ? ArrayUtil.median(weights, 0, n) : 0.0;
		
		if(n == 0) { 
			comments += "(DOF?)";
			return;
		}
		
		double maxWeight = max * medianWeight;
		double minWeight = min * medianWeight;
		
		double sumw = 0.0;
		n=0;
		for(int c=0; c<channels; c++) if(!channel[c].isBad) if(channel[c].weight > 0.0) {
			double wg2 = channel[c].weight * channel[c].gain * channel[c].gain;
			
			if(wg2 > maxWeight || wg2 < minWeight) channel[c].flag |= LabocaPixel.FLAG_SENSITIVITY;
			
			else {
				channel[c].flag &= ~LabocaPixel.FLAG_SENSITIVITY;
				if(channel[c].flag == 0) {
					sumw += wg2;
					n++;
				}
			}
		}
		if(n <= 0) { 
			comments += "(NEFD?)";
			return;
		}
				
		nefd = Math.sqrt(n*integrationTime/sumw)/(Unit.uV * Math.sqrt(Unit.s)) * janskyPerMicroV;
		
		census();
		comments += "(" + Util.e2.format(nefd) + ")";
		comments += goodChannelCount + " ";
	}

	public void despikeNeighbouring(double significance) {
		comments += "dN";

		for(int c=0; c<channels; c++) {
			double point =Double.NaN;
			if(frame[0] != null && (frame[0].flag & LabocaFrame.MODELING_FLAGS) == 0) 
				if((frame[0].sampleFlag[c] & LabocaFrame.SAMPLE_SOURCE_BLANK) == 0) point = frame[0].data[c];
			
			double next = point;

			double level = significance / Math.sqrt(0.5 * channel[c].weight);

			if(frame[0] != null) frame[0].sampleFlag[c] &= ~LabocaFrame.SAMPLE_SPIKY_NEIGHBOUR;
			
			for(int t=0; t < frames-1; t++) {
				next = Double.NaN;

				LabocaFrame exposure = frame[t];
				LabocaFrame nextFrame = frame[t+1];
				if(nextFrame != null) {
					nextFrame.sampleFlag[c] &= ~LabocaFrame.SAMPLE_SPIKY_NEIGHBOUR;
					if((nextFrame.flag & LabocaFrame.MODELING_FLAGS) == 0)
						if((nextFrame.sampleFlag[c] & LabocaFrame.SAMPLE_SOURCE_BLANK) == 0) next = nextFrame.data[c];
				}	
					
				
				if(exposure != null) {
					if(Math.abs(point - next) > level) {
						exposure.sampleFlag[c] |= LabocaFrame.SAMPLE_SPIKY_NEIGHBOUR;
						nextFrame.sampleFlag[c] |= LabocaFrame.SAMPLE_SPIKY_NEIGHBOUR;
					}
				}

				point = next;
			}
		}

	}

	public void despikeAbsolute(double significance) {
		comments += "dA";

		for(int c=0; c<channels; c++) {
			double level = significance / Math.sqrt(channel[c].weight);

			for(int t=0; t < frames; t++) if(frame[t] != null) 
				if((frame[t].sampleFlag[c] & LabocaFrame.SAMPLE_SOURCE_BLANK) == 0) {
					if(Math.abs(frame[t].data[c]) > level) frame[t].sampleFlag[c] |= LabocaFrame.SAMPLE_SPIKE;
					else frame[t].sampleFlag[c] &= ~LabocaFrame.SAMPLE_SPIKE;
				}
		}
	}
	
	public void despikeGradual(double significance, double depth) {
		double[] channelRMS = new double[channels];
		
		for(int c=0; c<channels; c++) {
			channelRMS[c] = Math.sqrt(1.0 / channel[c].weight);
		}
			
		comments += "dG";
		
		for(int t=0; t < frames; t++) if(frame[t] != null) if((frame[t].flag & LabocaFrame.MODELING_FLAGS) == 0) {
			LabocaFrame exposure = frame[t];
			
			double maxdev = 0.0;
			// Find the largest not yet flagged as spike deviation.
			for(int c=0; c<channels; c++) if(channel[c].flag == 0) if((exposure.sampleFlag[c] & LabocaFrame.SAMPLE_SOURCE_BLANK) == 0) {
				double dev = Math.abs(exposure.data[c] / channel[c].gain);
				if(dev > maxdev) maxdev = dev;
			}
			if(maxdev > 0.0) {
				double minSignal = depth * maxdev;
				for(int c=0; c<channels; c++) if((exposure.sampleFlag[c] & LabocaFrame.SAMPLE_SOURCE_BLANK) == 0) {
					double level = Math.max(significance * channelRMS[c], channel[c].gain * minSignal);
					if(Math.abs(exposure.data[c]) > level) exposure.sampleFlag[c] |= LabocaFrame.SAMPLE_SPIKE;
					else exposure.sampleFlag[c] &= ~LabocaFrame.SAMPLE_SPIKE;
				}
			}
		}
		
	}
	
	

	


	public void despikeFeatures(double significance) {
		despikeFeatures(significance, driftN);
	}
	
	public void despikeFeatures(double significance, int maxBlockSize) {
		if(maxBlockSize > frames) maxBlockSize = frames;
		comments += "dF" + maxBlockSize;
		
		WeightedPoint[] timeStream = new WeightedPoint[frames];
		for(int t=0; t<frames; t++) timeStream[t] = new WeightedPoint();		
		
		WeightedPoint diff = new WeightedPoint();
		
		for(int c=0; c<channels; c++) {	
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				WeightedPoint point = timeStream[t];
				point.weight = 0.0;
				
				if(exposure != null) {
					point.value = exposure.data[c];
					exposure.sampleFlag[c] &= ~LabocaFrame.SAMPLE_SPIKY_FEATURE; // Clear the feature spike flag...
					if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0) point.weight = channel[c].weight;					
				}
			}
				
			int n = frames;
			// check and divide...
			
			for(int blockSize = 1; blockSize <= maxBlockSize; blockSize *= 2) {
				for(int T=1; T < n; T++) if(T < n) {
					diff.set(timeStream[T]);
					diff.subtract(timeStream[T-1]);
					if(Math.abs(diff.value) * Math.sqrt(diff.weight) > significance) {
						for(int t=blockSize*(T-1), blockt=0; t<frames && blockt < 2*blockSize; t++, blockt++) {
							LabocaFrame exposure = frame[t];
							if(exposure != null) exposure.sampleFlag[c] |= LabocaFrame.SAMPLE_SPIKY_FEATURE;		
						}
					}
				}
				
				n /= 2;
				
				for(int to=0, from=0; to < n; to++, from += 2) {
					timeStream[to] = timeStream[from];
					timeStream[to].average(timeStream[from+1]);
				}
			}
		}	

	}
	
	public void flagSpikyChannels(int spikeTypes, double flagFraction, int minSpikes, int channelFlag) {
		int maxChannelSpikes = Math.max(minSpikes, (int)Math.round(flagFraction * frames));
		
		// Flag spiky channels even if spikes are in spiky frames
		//int frameFlags = LabocaFrame.MODELING_FLAGS & ~LabocaFrame.FLAG_SPIKY;
		
		// Only flag spiky channels if spikes are not in spiky frames
		int frameFlags = LabocaFrame.MODELING_FLAGS;
		
		for(int c=0; c<channels; c++) {
			channel[c].spikes = 0;
			
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & frameFlags) == 0) if((exposure.sampleFlag[c] & spikeTypes) != 0) channel[c].spikes++;
			}

			if(channel[c].spikes > maxChannelSpikes) channel[c].flag |= channelFlag;
			else channel[c].flag &= ~channelFlag;
		}
		
		census();
		comments += goodChannelCount + " ";
	}
			
	public void flagSpikyFrames(int spikeTypes, double minSpikes) {
		int spikyFrames = 0;
		
		// Flag spiky frames even if spikes are in spiky channels.
		//int channelFlags = ~(LabocaPixel.FLAG_SPIKY | LabocaPixel.FLAG_FEATURES);
		
		// Flag spiky frames only if spikes are not in spiky channels.
		int channelFlags = ~0;
		
		for(int t=0; t<frames; t++) {
			LabocaFrame exposure = frame[t];

			if(frame[t] != null) {
				int frameSpikes = 0;

				for(int c=0; c<channels; c++) if((channel[c].flag & channelFlags) == 0)
					if((exposure.sampleFlag[c] & spikeTypes) != 0) frameSpikes++;
				

				if(frameSpikes > minSpikes) {
					exposure.flag |= LabocaFrame.FLAG_SPIKY;
					spikyFrames++;
				}
				else exposure.flag &= ~LabocaFrame.FLAG_SPIKY;
			}
		}
		
		comments += "(" + Util.f1.format(100.0*spikyFrames/frames) + "%)";
	}
	
	
	
	

	
	
	
	
	
	public SkyMap getSourceIncrement(double minIntTime, boolean mapFilterCorrection) {
		comments += " [Map] ";
		
		int activeChannels = 0;
		for(int c=0; c<channels; c++) if(!channel[c].isBad) if(channel[c].flag == 0) activeChannels++;
		
		SkyMap scanMap = map.getSkyMap();
		scanMap.reset();

		// If there isn't enough good pixels in the scan, do not generate a map...
		// Require at least 50% of the total channels to work...
		if(activeChannels < 0.5*allChannels.length) return scanMap;

		calcAverageFiltering();
		
		// Only apply filtering corrections at the first source generation...
		if(map.generation == 0) comments += "[C1=" + Util.f2.format(1.0/averageFiltering) + "] ";
		else resetFiltering();
		
		// Normally the filtering of models is accounted for in the weights, which are fair weights 
		// for rederiving models -- i.e. they refer to what the true underlying rms was before modeling
		// However, in case of the source map, the signals are also filtered because of their interference 
		// with correlated noise. Therefore, when one scales the signals up to their unfiltered values,
		// Therefore it is the scaled fluxes that are to be added with fair weights. There is no need
		// to gain scale the weights for the filter correction...
		
		int goodFrames = 0;
		
		for(int t=0; t < frames; t++) {
			final LabocaFrame exposure = frame[t];
			if(exposure != null) if((exposure.flag & LabocaFrame.SOURCE_FLAGS) == 0) {

				final float[] residual = exposure.data;
				final double frameGain = exposure.transmission; 
				goodFrames++;

				if(frameGain > 0.0F) for(int k=0; k<allChannels.length; k++) {
					final LabocaPixel pixel = channel[allChannels[k]];
					
					if(pixel.flag == 0 && (exposure.sampleFlag[pixel.index] & LabocaFrame.SAMPLE_SPIKE_FLAGS) == 0) {
						final short[] index = getMapIndexes(t, pixel.index);
						final double g = pixel.mainBeamEfficiency * pixel.gain * frameGain * pixel.filtering;
						final double wg2 = g * g * pixel.weight;		
						if(g != 0.0) scanMap.addPointAtIndex(index[0], index[1], residual[pixel.index]/(g*pixel.filtering), wg2, integrationTime);
					}
				}
			}
		}
		
		scanMap.normalize();
		scanMap.directExposureClip(minIntTime);
		if(sourceGeneration == 0) scanMap.robustLevel();
		
		if(map.generation > 0 && mapFilterCorrection) {
			comments += "[C2=" + Util.f2.format(1.0/averageFiltering) + "] ";
			scanMap.applyCorrection(averageFiltering, map.s2n);
		}
		
		double sumw = 0.0;
		for(int k=0; k<allChannels.length; k++) if(channel[allChannels[k]].flag == 0) {
			LabocaPixel pixel = channel[allChannels[k]];
			sumw += pixel.weight * pixel.gain * pixel.gain;
		}
		
		double n = sumw > 0.0 ? (float) scanMap.countPoints() / sumw : 0;
		
		for(int k=0; k<allChannels.length; k++) {
			LabocaPixel pixel = channel[allChannels[k]];
			double G = pixel.gain * pixel.filtering * channel[pixel.index].mainBeamEfficiency;
			pixel.dependents += n * pixel.weight * G * G;
		}

		scanMap.integrationTime += goodFrames * integrationTime;
		
		return scanMap;

	}
	
	/*
	public void addSource() {	
	for(int t=0; t < frames; t++) if(frame[t] != null) if(!frame[t].flagged) {
	    final float[] residual = frame[t].data;
	    final double frameGain = frame[t].transmission; 
	
	    for(int bol=0; bol < channels; bol++) if(!LabocaPixel.isBad[bol]) {
		final double g = channel[bol].mainBeamEfficiency * channel[bol].gain * frameGain;
		final short[] index = getMapIndexes(t, bol);
		if(reduced.flag[index[0]][index[1]] == 0)
		    residual[bol] += g * (reduced.clean[index[0]][index[1]]);
	    }		    
	}
	}
	 */
	
	
	public SkyMap getSourceIncrement(int bol) {
		SkyMap scanMap = map.getSkyMap();
		scanMap.reset();
	
		if(channel[bol].flag != 0) return scanMap;
	
		final double dt = integrationTime;
	
		for(int t=0; t < frames; t++) if(frame[t] != null) if((frame[t].flag & LabocaFrame.SOURCE_FLAGS) == 0 && (frame[t].sampleFlag[bol] & LabocaFrame.SAMPLE_SPIKE_FLAGS) == 0) {
			final float[] residual = frame[t].data;
			final double frameGain = frame[t].transmission; 
	
			if(frameGain > 0.0F) {
				final short[] index = getMapIndexes(t, bol);
				final double g = channel[bol].mainBeamEfficiency * channel[bol].gain * frameGain;
				final double w = g * g * channel[bol].weight;		
				if(g != 0.0) scanMap.addPointAtIndex(index[0], index[1], residual[bol]/g, w, dt);
			}		    
		}
	
		scanMap.normalize();
	
		return scanMap;
	}

	public void removeSource(SkyMap source, double[][] old, double blankingLevel) {
		
		for(int t=0; t < frames; t++) {
			final LabocaFrame exposure = frame[t];
			if(exposure != null) if((exposure.flag & LabocaFrame.SKIP_SOURCE) == 0) {
				final float[] residual = frame[t].data;
				final double frameGain = frame[t].transmission; 

				for(int bol=0; bol < channels; bol++) if(!channel[bol].isBad) {
					final double g = channel[bol].mainBeamEfficiency * channel[bol].gain * frameGain;
					final short[] index = getMapIndexes(t, bol);

					// Do not check for flags, to get a true difference image...
					residual[bol] -= (float) g * (source.signal[index[0]][index[1]] - old[index[0]][index[1]]);
					
					// Do the blanking here...
					if(source.s2n[index[0]][index[1]] > blankingLevel) exposure.sampleFlag[bol] |= LabocaFrame.SAMPLE_SOURCE_BLANK;
					else exposure.sampleFlag[bol] &= ~LabocaFrame.SAMPLE_SOURCE_BLANK;
				}	    
			}
		}
		sourceGeneration++;
	}

	public void blankSource(Vector2D center, double flagRadius, double onSourceRadius) {
		blankSource(new Vector2D[] { center }, flagRadius, onSourceRadius);	
	}
	
	public void blankSource(Vector2D[] center, double flagRadius, double onSourceRadius) {
		//int lastSourcePixel = -1;
		
		for(int t=0; t < frames; t++) if(frame[t] != null) {
			for(int bol=0; bol < channels; bol++) if(!channel[bol].isBad) {
				Vector2D offset = frame[t].getdAZdEL(channel[bol]);
				for(int i=0; i<center.length; i++) {
					if(offset.distanceTo(center[i]) <= flagRadius) {
						if(offset.distanceTo(center[i]) <= onSourceRadius) 
							frame[t].sampleFlag[bol] |= LabocaFrame.SAMPLE_SOURCE_BLANK;
						else 
							frame[t].sampleFlag[bol] |= LabocaFrame.SAMPLE_SKIP;
					}
					/*
					if(frame[t].flag == 0) if(channel[bol].flag == 0) if(bol != lastSourcePixel) {
						if(offset.length() < radius) System.err.println("Found source in pixel " + (channel[bol].backendIndex + 1));
						lastSourcePixel = bol;
					}
					*/
				}
			}
		}	    
	}

	
	public WeightedPoint[] getSourceFluxes(Vector2D[] center, double radius) {
		double r2 = radius * radius;
		WeightedPoint[] fluxes = new WeightedPoint[center.length];
		for(int i=0; i<fluxes.length; i++) fluxes[i] = new WeightedPoint();
		
		for(int t=0; t < frames; t++) if(frame[t] != null) if((frame[t].flag & LabocaFrame.SOURCE_FLAGS) == 0) {
			for(int bol=0; bol < channels; bol++) if(!channel[bol].isBad) {
				final Vector2D offset = frame[t].getdAZdEL(channel[bol]);
				for(int i=0; i<center.length; i++) {
					offset.subtract(center[i]);
					if(offset.norm() <= r2) {
						int x = map.indexOfdX(offset.x);
						int y = map.indexOfdY(offset.y);
						double w = map.weight[x][y];
						
						fluxes[i].value += w * map.signal[x][y];
						fluxes[i].weight += w;						
					}
				}
			}	    
		}
		
		for(int i=0; i<fluxes.length; i++) fluxes[i].value /= fluxes[i].weight;
		
		return fluxes;
	}


	/*
    public void addSource() {	
	for(int t=0; t < frames; t++) if(frame[t] != null) if(!frame[t].flagged) {
	    final float[] residual = frame[t].data;
	    final double frameGain = frame[t].transmission; 

	    for(int bol=0; bol < channels; bol++) if(!LabocaPixel.isBad[bol]) {
		final double g = channel[bol].mainBeamEfficiency * channel[bol].gain * frameGain;
		final short[] index = getMapIndexes(t, bol);
		if(reduced.flag[index[0]][index[1]] == 0)
		    residual[bol] += g * (reduced.clean[index[0]][index[1]]);
	    }		    
	}
    }
	 */

	
	public int getNodPhase() { return onPhase == 1 ? 1 : -1; }

	/*
	public void flagChoppingPixels() {
		double halfBeam = 9.0 * Unit.arcsec;
		
		Vector2D on = new Vector2D();
		Vector2D left = new Vector2D(-2.0 * chopperThrow, 0.0);
		Vector2D right = new Vector2D(2.0 * chopperThrow, 0.0);
		
		for(int c=0; c<channels; c++) {
			Vector2D offset = channel[c].position;
			if(offset.distanceTo(on) < halfBeam) channel[c].flag |= LabocaPixel.FLAG_SKIP;
			else if(offset.distanceTo(left) < halfBeam) channel[c].flag |= LabocaPixel.FLAG_SKIP;
			else if(offset.distanceTo(right) < halfBeam) channel[c].flag |= LabocaPixel.FLAG_SKIP;
		}
		
		for(int c=0; c<channels; c++) if((channel[c].flag & LabocaPixel.FLAG_SKIP) != 0) {
			System.err.println(" Chopping on pixel " + (channel[c].backendIndex + 1));
			for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].sampleFlag[c] |= LabocaFrame.SAMPLE_SOURCE_BLANK;
		}
	}
	*/
	
	public void markChopped() {
		// Flag pixels that chop on source
		// left and right are the pixel positions, where, if there's a pixel, it will have the source
		// in the left or right beams...
		//System.err.println("on phase is " + subscan[i][k].onPhase);
		Vector2D left = new Vector2D(onPhase == 1 ? 0.0 : 2.0 * chopperThrow, 0.0);
		Vector2D right = new Vector2D(onPhase == 1 ? -2.0 * chopperThrow : 0.0, 0.0);

		for(int c=0; c<channels; c++) {
			LabocaPixel pixel = channel[c];
			if(pixel.position.distanceTo(left) < beamFWHM / 2.0) pixel.flag |= LabocaPixel.CHOPPED_LEFT_BEAM;
			else if(pixel.position.distanceTo(right) < beamFWHM / 2.0) pixel.flag |= LabocaPixel.CHOPPED_RIGHT_BEAM;
			if((pixel.flag & LabocaPixel.CHOPPED_BEAM) != 0) 
				System.err.println("   Source is chopped on pixel " + (pixel.backendIndex + 1));
		}
	}	

	
	public void centerChopped(double tolerance, boolean isRobust) {
		comments += "o";
		
		int fromt = 0;
		int n=0;
		
		while(fromt < frames) {
			fromt = centerChopped(fromt, tolerance, isRobust);	
			n++;
		}
		
		//System.err.println(n + " chopping cycles found.");
	}
	
	// Find center: (aveleft+averight)/2
	public int centerChopped(int fromt, double tolerance, boolean isRobust) {
		
		double[] left = new double[channels];
		double[] right = new double[channels];
		int nLeft=0, nRight=0;
		
		
		int sign = Math.abs(frame[fromt].chopperPosition) - tolerance > 0.0 ? (int)Math.signum(frame[fromt].chopperPosition) : 0; 
		boolean transitioned = false;
		boolean completedCycle = false;
		
		if(sign < 0) {
			for(int c=0; c<channels; c++) left[c] += frame[fromt].data[c];
			nLeft++;
		}
		else if(sign > 0) {
			for(int c=0; c<channels; c++) right[c] += frame[fromt].data[c];
			nRight++;
		}
		
		
		int tot=fromt;
		while(tot < frames && !completedCycle) {
			while(frame[tot] == null && tot < frames-1) tot++;
			int nextSign = Math.abs(frame[tot].chopperPosition) - tolerance > 0.0 ? (int)Math.signum(frame[tot].chopperPosition) : 0; 
			boolean isTransition = sign * nextSign == -1;
			if(isTransition) {
				if(!transitioned) transitioned = true;
				else completedCycle = true;
			}
			if(nextSign != 0) sign = nextSign;
			
			if(sign < 0) {
				for(int c=0; c<channels; c++) left[c] += frame[tot].data[c];
				nLeft++;
			}
			else if(sign > 0) {
				for(int c=0; c<channels; c++) right[c] += frame[tot].data[c];
				nRight++;
			}
			
			tot++;
		}
		
		if(!completedCycle) {
			for(int t=fromt; t<tot; t++) if(frame[t] != null) frame[t].flag |= LabocaFrame.BAD_DATA;
			return tot;
		}
		
		for(int c=0; c<channels; c++) {
			double value = (left[c]/nLeft + right[c]/nRight)/2.0;
			for(int t=fromt; t<tot; t++) if(frame[t] != null) frame[t].data[c] -= value;
		}
		
		return tot;		
	}

	public WeightedPoint getPhotometry() {
		double sum = 0.0, sumw = 0.0;
		int n = 0;
		for(int t=0; t<frames; t++) {
			LabocaFrame exposure = frame[t];
			if(exposure != null) if((exposure.flag & LabocaFrame.SOURCE_FLAGS) == 0)
				for(int c=0; c<channels; c++) if(channel[c].flag == 0) if(exposure.sampleFlag[c] == LabocaFrame.SAMPLE_SOURCE_BLANK) {
					LabocaPixel pixel = channel[c];
					double G = pixel.gain * pixel.mainBeamEfficiency * exposure.transmission;	
					sum += pixel.weight * G * exposure.data[c]; 
					sumw += pixel.weight * G * G;
					n++;
				}
		}
		
		return new WeightedPoint(sum/sumw, sumw);
	}
	
	public void shapeSpectrum(int windowSize, int driftT) {
		double[] shape = new double[windowSize/2 + 1];
		double sigmaf = 1.25 * shape.length/2.0;
		for(int f=1; f<shape.length; f++) {
			double chi = (float)f / sigmaf;
			shape[f] = Math.exp(-chi*chi/2.0);
			//shape[f] = 1.0;
		}

		int minF = (int)Math.ceil((float)windowSize / driftT);

		shapeSpectrum(windowSize, shape, minF);

	}

	public void shapeSpectrum(int windowSize, double[] shape, int minF) {
		if(frames < windowSize) {
			System.err.println("Whitening window too large.");
			return; 
		}

		int n = ArrayUtil.getPaddedFFTSize(frames);
		Complex[] spectrum = new Complex[n/2+1];
		double[] data = new double[n];

		for(int f=0; f<spectrum.length; f++) spectrum[f] = new Complex();

		for(int c=0; c<channels; c++) {
			
			Arrays.fill(data, 0.0);
			
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.BAD_DATA) == 0) if(exposure.sampleFlag[c] == 0)
					data[t] = frame[t].data[c];
			}	

			ArrayUtil.uncheckedFFT(data, spectrum);
			double[] P = ArrayUtil.averagePower(data, windowSize);
			double[] C = new double[P.length + 1];
			int stretch = n/windowSize;

			double aveP = 0.0, norm = 0.0;
			int from = P.length / 8;
			int to = P.length / 2;

			if(from < minF) from=minF;

			for(int f=from; f<to; f++) {
				aveP += P[f];
				norm += shape[f];
			}
			aveP /= norm;	    

			for(int f=0; f<P.length; f++) C[f] = Math.sqrt(P[f] / (shape[f] * aveP));

			for(int f=minF; f<spectrum.length; f++) {
				int F = (f/stretch);
				//double a = (float)f / stretch - (F-1);
				//double P0 = a * C[F] + (1.0 - a) * C[F+1];
				double P0 = C[F];
				spectrum[f].scale(1.0/P0);
			}

			ArrayUtil.uncheckedBackFFT(spectrum, data);

			for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].data[c] = (float) data[t];
		}


	}
	
	public void whitenToMedian(double level) {
		whitenToMedian(ArrayUtil.getPaddedFFTSize(2 * driftN), level);
	}
		
	public void whitenToMedian(int windowSize, double level) {
		comments += "[whiten]";
		
		windowSize = ArrayUtil.getPaddedFFTSize(windowSize);
		
		if(frames < windowSize) {
			System.err.println("Whitening window too large.");
			return; 
		}

		int n = ArrayUtil.getPaddedFFTSize(frames);
		Complex[] spectrum = new Complex[n/2+1];
		double[] data = new double[n];
		int stretch = n/windowSize;
		
		for(int f=0; f<spectrum.length; f++) spectrum[f] = new Complex();
		
		double[] amplitude = new double[1 + windowSize / 2];
		double[] scaling = new double[1 + windowSize / 2];
		int[] nAmp = new int[1 + windowSize / 2];
		
		int pixelFlags = LabocaPixel.FLAG_DEAD | LabocaPixel.FLAG_BLIND | LabocaPixel.FLAG_DISCONTINUITY | LabocaPixel.FLAG_GAIN;
		
		
		for(int c=0; c<channels; c++) if((channel[c].flag & pixelFlags) == 0) {
			
			Arrays.fill(data, 0.0);
	
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure != null) if((exposure.flag & LabocaFrame.MODELING_FLAGS) == 0) if(exposure.sampleFlag[c] == 0)
					data[t] = frame[t].data[c];
			}

			ArrayUtil.uncheckedFFT(data, spectrum);

			Arrays.fill(amplitude, 0.0);
			Arrays.fill(nAmp, 0);
			
			for(int f=1; f<spectrum.length; f++) {
				int F = f/stretch;
				amplitude[F] += spectrum[f].length();
				nAmp[F]++;
			}
			
			
			for(int f=0; f<amplitude.length; f++) amplitude[f] /= nAmp[f];
				
			int nWindows = frames/windowSize;
			double aveAmp = ArrayUtil.median(amplitude, 1, amplitude.length);
			double sigmaA = aveAmp / Math.sqrt(nWindows);
			double critical = aveAmp * level + 3.0*sigmaA;
			
			// Only whiten those frequencies which have a significant excess power
			// when compared to the specified level over the median spectral power.
			for(int f=0; f<scaling.length; f++) scaling[f] = amplitude[f] > critical ?  aveAmp / amplitude[f] : 1.0;
			
			// Renormalize the whitening scaling s.t. it is median-power neutral
			double norm = ArrayUtil.median(scaling);
			double aveScaling = 0.0; 
			for(int f=1; f<amplitude.length; f++) {
				scaling[f] /= norm;
				aveScaling += scaling[f];
			}
			aveScaling /= amplitude.length - 1;
			channel[c].whitening *= aveScaling;
			channel[c].filtering *= channel[c].whitening;
			
			for(int f=1; f<spectrum.length; f++) spectrum[f].scale(scaling[f/stretch]);
	
			ArrayUtil.uncheckedBackFFT(spectrum, data);

			for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].data[c] = (float) data[t];
		}


	}

	public String getFileName(String path, int scanNumber, String projectID) {		
		File root = new File(path);
		String[] fileName = root.list();
		for(int i=0; i<fileName.length; i++) {
			if(fileName[i].startsWith(projectID)) 
				return getFileName(path + File.separator + fileName[i], scanNumber, projectID);
			if(fileName[i].startsWith("APEX-" + scanNumber)) {
				if(fileName[i].endsWith(".gz")) fileName[i] = fileName[i].substring(0, fileName[i].length()-3);
				if(fileName[i].endsWith(".fits")) fileName[i] = fileName[i].substring(0, fileName[i].length()-5);
				if(fileName[i].endsWith(projectID))
					return path + File.separator + fileName[i];
			}
		}
		return null;
	}
	

	public static LabocaScan unite(LabocaScan[] subscans) {
		LabocaScan scan = (LabocaScan) subscans[0].clone();
		
		Vector<LabocaFrame> buffer = new Vector<LabocaFrame>();
		for(int i=0; i<subscans.length; i++) {
			LabocaScan subscan = subscans[i];
			subscan.trim();

			if(i>0) {
				double mjdGap = subscan.frame[0].MJD - subscans[i-1].frame[subscans[i-1].frames-1].MJD;
				int padding = (int) Math.round(mjdGap * Unit.day / scan.integrationTime);
				if(padding < 0) padding = 0;
				for(int t=0; t<padding; t++) buffer.add(null);
			}

			for(int t=0; t<subscan.frames; t++) buffer.add(subscan.frame[t]);

		}

		scan.frames = buffer.size();
		scan.frame = buffer.toArray(new LabocaFrame[scan.frames]);
		
		return scan;
	}
	
	public LabocaScan readAsOne(int scanNumber, String projectID, boolean gzipped) throws IOException, FitsException, HeaderCardException {
		return unite(read(scanNumber, projectID, gzipped));
	}
		
	public LabocaScan[] read(int scanNumber, String projectID, boolean gzipped) throws IOException, FitsException, HeaderCardException {
		LabocaScan common = (LabocaScan) clone();
		common.serialNo = scanNumber;
		common.isCompressed = gzipped;
		//common.descriptor = "APEX-" + scanNumber + "-" + projectID

		System.err.println();
		System.err.println("Scan " + scanNumber + " (details follow...)");

		//System.err.println("Looking for .fits" + (gzipped ? ".gz" : " files") + "\n   with project ID " + projectID + "\n   in " + path);
		String name = getFileName(path, scanNumber, projectID);
		String ext = ".fits" + (gzipped ? ".gz" : "");
		
		File file = new File(name);
		if(file.isDirectory()) {	
			//System.err.println("Found scan directory " + name); 		
			return readScanDirectory(common, name, ext);
		}
		else {
			//System.err.println("Found scan " + name);
			return readScan(common, name + ext);
		}
	}
	

	
	public LabocaScan[] readScanDirectory(LabocaScan common, String dir, String ext) throws IOException, FitsException, HeaderCardException {
		int subscans = common.readScanInfo(dir + File.separator + "SCAN" + ext);
		common.readPar(dir + File.separator + "LABOCA-ABBA-FEBEPAR" + ext);

		Vector<LabocaScan> buffer = new Vector<LabocaScan>();

		for(int i=0; i<subscans; i++) {
			try {
				LabocaScan subscan = (LabocaScan) common.copy();

				System.err.println(" Subscan {" + (i+1) + "}:");
				subscan.subscanNo = i;
				subscan.readDataPar(dir + File.separator + (i+1) + File.separator + "LABOCA-ABBA-DATAPAR" + ext);
				subscan.readData(dir + File.separator + (i+1) + File.separator + "LABOCA-ABBA-ARRAYDATA-1" + ext);
				subscan.readMonitor(dir + File.separator + (i+1) + File.separator + "MONITOR" + ext);
				subscan.validate();

				buffer.add(subscan);
			}
			catch(Exception e) {
				System.err.println(" ERROR: " + e.getMessage());
				e.printStackTrace();
			}
		}

		LabocaScan[] set = new LabocaScan[buffer.size()];
		for(int i=0; i<buffer.size(); i++) {
			set[i] = buffer.get(i);
			set[i].scanPath = dir;
		}
		
		return set;
	}
	
	public LabocaScan[] readScan(LabocaScan common, String fileName) throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));	
		BasicHDU[] hdu = fits.read();
		
		int subscans = common.readScanInfo((BinaryTableHDU) hdu[1]);
		common.readPar((BinaryTableHDU) hdu[2]);

		Vector<LabocaScan> buffer = new Vector<LabocaScan>();

		int k=3;
		for(int i=0; i<subscans; i++) {
			try {
				LabocaScan subscan = (LabocaScan) common.copy();
				subscan.subscanNo = i;
				
				System.err.println(" Subscan {" + (i+1) + "}:");
				
				// HDUs for each subscan can come in any order, so check EXTNAME...
				for(int m=0; m<3; m++, k++) {
					BinaryTableHDU table = (BinaryTableHDU) hdu[k];
					String extName = table.getHeader().getStringValue("EXTNAME");
					
					if(extName.equalsIgnoreCase("DATAPAR-MBFITS")) subscan.readDataPar(table);
					else if(extName.equalsIgnoreCase("ARRAYDATA-MBFITS")) subscan.readData(table);
					else if(extName.equalsIgnoreCase("MONITOR-MBFITS")) subscan.readMonitor(table);
				}
				
				subscan.validate();

				buffer.add(subscan);
			}
			catch(Exception e) {
				System.err.println(" ERROR: " + e.getMessage());
				e.printStackTrace();
			}
		}

		LabocaScan[] set = new LabocaScan[buffer.size()];
		for(int i=0; i<buffer.size(); i++) set[i] = buffer.get(i);

		return set;
	}
	

	// returns the number of subscans.
	public int readScanInfo(String fileName) throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));	

		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
		return readScanInfo(hdu);
	}
	
	public int readScanInfo(BinaryTableHDU hdu) throws IOException, FitsException, HeaderCardException {
		Header header = hdu.getHeader();
		// Read in the full HDU data
		//hdu.getData().getData();
		
		equatorial = new EquatorialCoordinates(header.getDoubleValue("CRVAL1") * Unit.deg, header.getDoubleValue("CRVAL2") * Unit.deg, "J2000");
		sourceName = header.getStringValue("OBJECT");

		System.err.println(" Source [" + sourceName + "]");

		if(sourceName == null) sourceName = "LABOCA";
		LabocaFrame.longitude = header.getDoubleValue("SITELONG") * Unit.deg;
		LabocaFrame.latitude = header.getDoubleValue("SITELAT") * Unit.deg;

		//isChopped = header.getBooleanValue("WOBUSED");
		chopperThrow = header.getDoubleValue("WOBTHROW") * Unit.deg;
		
		isNativeEquatorial = header.getStringValue("CTYPE1").startsWith("RA");

		return header.getIntValue("NOBS");	
	}

	public void readPar(String fileName) throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));

		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
		readPar(hdu);
	}
	
	public void readPar(BinaryTableHDU hdu) throws IOException, FitsException, HeaderCardException {
		Header header = hdu.getHeader();

		G = 1.0 / (270 * 1<<(int)header.getDoubleValue("FEGAIN"));
		//System.err.println(" Frontend Gain is " + (1.0/G));

		Object[] row = hdu.getRow(0);

		double[] xOffset = (double[]) row[5];
		double[] yOffset = (double[]) row[6];
		referenceChannel = ((int[]) row[7])[0]-1;

		//referenceChannel = 12;

		System.err.println(" Reference pixel is " + (referenceChannel+1));

		channels = xOffset.length;
		channel = new LabocaPixel[channels];

		System.err.println(" " + channels + " channels found.");

		// LabocaPixel.offset: get BOLDCOFF

		for(int c=0; c<channels; c++) {
			channel[c] = (LabocaPixel) channelTemplate.clone();
			channel[c].index = c;
			channel[c].backendIndex = c;
			channel[c].position = new Vector2D(xOffset[c] * Unit.deg, yOffset[c] * Unit.deg);
			channel[c].position = new Vector2D();
		}
		
		LabocaPixel.setReference(channel, channel[referenceChannel]);

		// Needed only for pre Feb2007 Files with integer format.
		//G *= 10.0/ ((float[]) row[17])[0] / (1<<15-1);	
	}

	boolean isCompressed = false;
	
	public void readData(String fileName) throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));

		BinaryTableHDU hdu =(BinaryTableHDU) fits.getHDU(1);
		readData(hdu);
	}
	
	public void readData(BinaryTableHDU hdu) throws IOException, FitsException, HeaderCardException {
		//Header header = hdu.getHeader();

		// Read in the full HDU data
		hdu.getData().getData();
		
		Vector<LabocaFrame> buffer = new Vector<LabocaFrame>();
		double lastMJD = Double.NaN;

		for(int t=0; t<frames; t++) {
			LabocaFrame exposure = frame[t];
			Object[] row = hdu.getRow(t);
			double MJD = ((double[]) row[0])[0] * Unit.day;

			if(!Double.isNaN(lastMJD)) {
				int n = (int)Math.round((MJD - lastMJD) / integrationTime);
				if(n > 50) {
					System.err.println("   Inserting " + n + " empty frames...");
					for(int i=0; i<n; i++) buffer.add(null);
				}
			}

			lastMJD = MJD;

			float[][] data = (float[][]) row[1];
			exposure.data = new float[data.length];

			for(int c=0; c<data.length; c++)
				exposure.data[c] = (float)G * data[c][0];

			exposure.calc();

			buffer.add(exposure);
		}	
		
		for(int c=0; c<channels; c++) {
			channel[c].dependents = 0;
			channel[c].spikes = 0;
		}

		frame = new LabocaFrame[frames];
		for(int t=0; t<frames; t++) frame[t] = buffer.get(t);


	}
	
	public void writeData(String fromName, String toName) throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fromName), fromName.endsWith(".gz"));
		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
			
		for(int t=0; t<frames; t++) {
			LabocaFrame exposure = frame[t];
			final Object[] row = hdu.getRow(t);
			final float[][] data = (float[][]) row[1];
			for(int c=0; c<data.length; c++) data[c][0] = 0.0F;
			for(int c=0; c<channels; c++) data[channel[c].backendIndex][0] = exposure.data[c];
			hdu.setRow(t, row);
		}	
			
		fits.write(new BufferedDataOutputStream(new FileOutputStream(toName)));
	}
	
	public void clearData() {
		for(int t=0; t<frames; t++) {
			final LabocaFrame exposure = frame[t];
			if(exposure != null) for(int c=0; c<channels; c++) exposure.data[c] = 0.0F;
		}
	}
	
	public void randomData() {
		Random random = new Random();
		
		float[] rms = new float[channels];
		for(int c=0; c<channels; c++) rms[c] = (float)(Math.sqrt(1.0/channel[c].weight) / G);
		
		for(int t=0; t<frames; t++) {
			final LabocaFrame exposure = frame[t];
			if(exposure != null) for(int c=0; c<channels; c++) if(!channel[c].isBad) if(channel[c].flag == 0)
				exposure.data[c] = rms[c] * (float)random.nextGaussian();						
		}	
	}
	
	// Adds correlated noise with 1/f^2 spectrum
	public void addCorrelated(int[] channelSelection, float[] gain, double nyquistLevel) {		
		double prior = 0.0;
		Random random = new Random();
		for(int t=0; t<frames; t++) {
			double signal = prior + nyquistLevel * random.nextGaussian();
			prior = signal;
			LabocaFrame exposure = frame[t];
			if(exposure != null) for(int k=0; k<channelSelection.length; k++) {
				final LabocaPixel pixel = channel[channelSelection[k]];
				if(!pixel.isBad) if(pixel.flag == 0) exposure.data[pixel.index] += gain[k] * signal / (float) G;
			}
		}	
	}
	

	// Peak is specified in volts at the detector...
	public void addSources(Vector<GaussianSource> sources) {
		double[] A = new double[sources.size()];
		Vector2D[] centerOffset = new Vector2D[sources.size()];
		
		for(int i=0; i<sources.size(); i++) {
			GaussianSource source = sources.get(i);
			double sigma = source.FWHM / 2.35;
			A[i] = -0.5 / (sigma * sigma);
			centerOffset[i] = new Vector2D(); 
			map.projection.project(source.coords, centerOffset[i]);
			
			System.err.print("  Source #" + (i+1) + " is at " + Util.f2.format(centerOffset[i].x/Unit.arcsec) + "," + Util.f2.format(centerOffset[i].y/Unit.arcsec));
			System.err.println(" peak = " + (source.peak / Unit.uV) + " uV");
		}
		
		for(int t=0; t < frames; t++) {
			final LabocaFrame exposure = frame[t];
			if(exposure != null) {
				final float[] residual = frame[t].data;
				final double frameGain = frame[t].transmission; 

				for(int bol=0; bol < channels; bol++) if(!channel[bol].isBad) {
					final double g = channel[bol].mainBeamEfficiency * channel[bol].gain * frameGain / G;
					final Vector2D offset = getMapOffsets(t, bol);
					
					for(int i=0; i<A.length; i++) {
						final GaussianSource source = sources.get(i);
						final double d = offset.distanceTo(centerOffset[i]);
					
						// Do not check for flags, to get a true difference image...
						residual[bol] += (float) (g * source.peak * Math.exp(A[i] * d * d));
					}
				}	    
			}
		}
	}
	
	
	


	public void readDataPar(String fileName)  throws IOException, FitsException, HeaderCardException {
		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));

		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
		readDataPar(hdu);
	}
	
	
	public void readDataPar(BinaryTableHDU hdu)  throws IOException, FitsException, HeaderCardException {
		Header header = hdu.getHeader();

		// Read in the full HDU data
		hdu.getData().getData();
		
		frames = header.getIntValue("NAXIS2");
		integrationTime = ((double[]) hdu.getRow(0)[hdu.findColumn("INTEGTIM")])[0];
		System.err.println("   Sampling at " + Util.f3.format(1.0/integrationTime) + " Hz.");

		int iMJD = hdu.findColumn("MJD");
		int iRA = hdu.findColumn("BASLONG");
		int iDEC = hdu.findColumn("BASLAT");
		int iAZ = hdu.findColumn("AZIMUTH");
		int iEL = hdu.findColumn("ELEVATIO");
		int iDX = hdu.findColumn("LONGOFF");
		int iDY = hdu.findColumn("LATOFF");
		int iChop = hdu.findColumn("WOBDISLN");
		int iPhase = hdu.findColumn("PHASE");

		System.err.println("   " + frames + " frames found (" + 
				Util.f1.format(frames * integrationTime / Unit.min) + " minutes).");


		frame = new LabocaFrame[frames];
		
		if(scramble) System.err.println("   !!! Scrambling position data (noise map only) !!!");
		
		for(int t=0, T=0; t<frame.length; t++) {
			LabocaFrame exposure = new LabocaFrame(this);
			Object[] row = hdu.getRow(T++);
			exposure.MJD = ((double[])row[iMJD])[0];
			exposure.RA = ((double[])row[iRA])[0] * Unit.deg;
			exposure.DEC = ((double[])row[iDEC])[0] * Unit.deg;
			exposure.AZ = ((double[])row[iAZ])[0] * Unit.deg;
			exposure.EL = ((double[])row[iEL])[0] * Unit.deg;
			exposure.dX = ((double[])row[iDX])[0] * Unit.deg;
			exposure.dY = ((double[])row[iDY])[0] * Unit.deg;
			if(iChop > 0) exposure.chopperPosition = ((double[])row[iChop])[0] * Unit.deg;
			exposure.chopperPhase = ((int[])row[iPhase])[0];	

			if(scramble) {
				exposure.dX *= -1.0;
				exposure.dY *= -1.0;
				exposure.RA = 2.0 * equatorial.RA() - exposure.RA;
				exposure.DEC = 2.0 * equatorial.DEC() - exposure.DEC;
			}
			
			exposure.zenithTau = zenithTau;
			
			//exposure.calc();
			
			frame[t] = exposure;
		}	


		String phase1 = header.getStringValue("PHASE1");	
		String phase2 = header.getStringValue("PHASE2");

		if(phase1 != null) {
			if(phase1.equalsIgnoreCase("WON")) onPhase = 1;
			else if(phase1.equalsIgnoreCase("WOFF")) offPhase = 1;
		}
		if(phase2 != null) {
			if(phase2.equalsIgnoreCase("WON")) onPhase = 2;
			if(phase2.equalsIgnoreCase("WOFF")) offPhase = 2;
		}

		if(isChopped) System.err.println("   " + (getNodPhase() > 0 ? "ON-OFF" : "OFF-ON"));

		horizontal = new HorizontalCoordinates(frame[frames/2].AZ, frame[frames/2].EL);
		PA = frame[frames/2].PA;
	}



	public void readMonitor(String fileName)  throws IOException, FitsException, HeaderCardException {
		if(!(useHe3Data || calcHe3Gains)) return;

		Fits fits = new Fits(new File(fileName), fileName.endsWith(".gz"));

		BinaryTableHDU hdu = (BinaryTableHDU) fits.getHDU(1);
		readMonitor(hdu);
	}
	
	public void readMonitor(BinaryTableHDU hdu)  throws IOException, FitsException, HeaderCardException {
		System.err.print("   Parsing/Interpolating He3 temperatures. ");
		Header header = hdu.getHeader();

		hdu.getData().getData();
		
		Vector<Vector2D> table = new Vector<Vector2D>();
		
		int iMJD = hdu.findColumn("MJD");
		int iLABEL = hdu.findColumn("MONPOINT");
		int iVALUE = hdu.findColumn("MONVALUE");
		
		int nFlagged = 0;

		int n = header.getIntValue("NAXIS2");
		for(int i=0; i<n; i++) {
			Object[] row = hdu.getRow(i);	
		
			if(((String) row[iLABEL]).equals("LABOCA_HE3TEMP")) {
				Vector2D entry = new Vector2D();
				entry.x = ((double[]) row[iMJD])[0];
				entry.y = ((double[]) row[iVALUE])[0];
				table.add(entry);
			}
		}

		int from = 0, to=0;

		// plus or minus in days;
		double smoothFWHM = he3TimeScale;
		double windowSize = 2.0 * smoothFWHM;

		double sigma = smoothFWHM / 2.35;

		for(int t=0; t<frames; t++) {
			double MJD0 = frame[t].MJD;

			// Adjust the smoothing window...
			while(MJD0 - table.get(from).x > windowSize || from == table.size()-1) from++;
			while(table.get(to).x - MJD0 < windowSize || to == table.size()-1) to++;

			double sum = 0.0, sumw = 0.0;

			for(int tt=from; tt<=to; tt++) {
				Vector2D entry = table.get(tt);
				double dev = (MJD0 - entry.x) / sigma;
				double w = Math.exp(-0.5*dev*dev);
				sum += w * entry.y;
				sumw += w;
			}

			frame[t].he3Temp = sum/sumw;
			if(Double.isNaN(frame[t].he3Temp)) {
				frame[t] = null;
				nFlagged++;
			}
		}


		System.err.println(nFlagged == 0 ? "All good :-)" : "Dropped " + nFlagged + " frames without reliable He3 data :-(");

	}


	public Vector2D[] getSmoothPosition(int n) {
		Vector2D[] pos = new Vector2D[frames];

		int empties = 0;
		double intX=0.0, intY=0.0;
		for(int t=0; t<n; t++) {
			if(frame[t] == null) empties++;
			else {	    
				intX += frame[t].dX;
				intY += frame[t].dY;
			}
		}

		int tot = frames-n/2-1;
		for(int t=n/2; t<tot; t++) {
			pos[t] = new Vector2D(intX/(n - empties), intY/ (n - empties));
			LabocaFrame oldFrame = frame[t-n/2];
			LabocaFrame newFrame = frame[t+n/2];

			if(newFrame == null) empties++;
			else {
				intX += newFrame.dX;
				intY += newFrame.dY;
			}

			if(oldFrame == null) empties--;
			else {
				intX -= oldFrame.dX;
				intY -= oldFrame.dY;
			}

		}

		pos[tot] = n > empties ? new Vector2D(intX/(n-empties), intY/(n-empties)) : null;

		return pos;
	}


	
	public double getAverageScanningVelocity(double smoothT) {
		int N = (int)Math.ceil(smoothT/integrationTime);
		Vector2D[] pos = getSmoothPosition(N);

		double[] v = new double[frames]; 
		int n = 0;
		
		int tot = frames-N/2-2;
		
		for(int t=N/2+1; t<tot; t++) {
			LabocaFrame exposure = frame[t];
			if(exposure != null) if((exposure.flag & LabocaFrame.SKIP_SOURCE) == 0) {
				double dX = (pos[t+1].x - pos[t-1].x);
				double dY = (pos[t+1].y - pos[t-1].y);
				v[n++] = Math.hypot(dX, dY) / (2.0 * integrationTime);
			}
		}
		
		return n > 0 ? ArrayUtil.median(v) : Double.NaN;
	}
	
	
	public int velocityCut(double minV, double maxV) { return velocityCut(minV, maxV, 0.5 * Unit.s); }
	
	
	public int velocityCut(double minV, double maxV, double smoothT) {
		Vector2D[] pos = getSmoothPosition((int)Math.ceil(smoothT/integrationTime));

		double minD = minV*2.0*integrationTime;	
		double maxD = maxV*2.0*integrationTime;	

		int cut = 2;

		for(int t=1; t<frames-1; t++) if(frame[t] != null) {
			if(pos[t+1] == null || pos[t-1] == null) {
				frame[t].flag |= LabocaFrame.SKIP_SOURCE;
				cut++;
			}
			else {
				double D = Math.hypot(pos[t+1].x - pos[t-1].x, pos[t+1].y - pos[t-1].y);
				
				if(D < minD || D > maxD) {
					frame[t].flag |= LabocaFrame.SKIP_SOURCE;
					cut++;
				}
			}
		}

		return cut;
	}

	public int accelerationCut(double maxA) {
		Vector2D[] pos = getSmoothPosition(10);

		double maxD2 = maxA*integrationTime*integrationTime;
		maxD2 *= maxD2;	

		int cut = 0;

		for(int t=1; t<frames-1; t++) if(frame[t] != null) {
			if(pos[t+1] == null || pos[t-1] == null) {
				frame[t] = null;
				cut++;
			}
			else {
				double dX = (pos[t+1].x + pos[t-1].x - 2.0*pos[t].x);
				double dY = (pos[t+1].y + pos[t-1].y - 2.0*pos[t].y);
				if((dX*dX+dY*dY) > maxD2) {
					frame[t] = null;
					cut++;
				}
			}
		}

		return cut;
	}

	public Vector2D[] getAccelerations() { return getAccelerations(0.5); }
	
	public Vector2D[] getAccelerations(double smoothT) {
		Vector2D[] pos = getSmoothPosition((int)Math.ceil(smoothT/integrationTime));
		Vector2D[] accel = new Vector2D[frames];

		for(int t=1; t<frames-1; t++) {
			if(pos[t+1] == null || pos[t-1] == null) accel[t] = null;
			else {
				accel[t] = new Vector2D();
				accel[t].x = (pos[t+1].x + pos[t-1].x - 2.0*pos[t].x);
				accel[t].y = (pos[t+1].y + pos[t-1].y - 2.0*pos[t].y);
			}
		}

		return accel;
	}

	public void pointingAt(Vector2D dAZdEL) {
		for(int c=0; c<channels; c++)
			channel[c].position.math('-', dAZdEL);
	}

	public void highPassFilter(double T) {
		comments += "H";
		
		int nFFT = ArrayUtil.getPaddedFFTSize(frames);
		int nf = nFFT/2+1;
		double[] signal = new double[nFFT];

		// sigmat sigmaw = 1 ->
		// Dt sigmaw = 2.35 
		// Dt 2Pi sigmaf = 2.35 --> sigmaf = 2.35/2Pi 1/Dt
		// df = 1/Nt;
		// sigmaf / df = 2.35/2Pi * Nt/Dt 
	
		double sigma_f = 2.35/(2.0*Math.PI) * frames * integrationTime / T;		
		double A = 0.5/(sigma_f*sigma_f);

		Complex[] spectrum = new Complex[nf];
		for(int f=0; f<nf; f++) spectrum[f] = new Complex();

		for(int bol=0; bol<channels; bol++) {
			for(int t=0; t<frames; t++) {
				LabocaFrame exposure = frame[t];
				if(exposure == null) signal[t] = 0.0;
				else if((exposure.flag & LabocaFrame.MODELING_FLAGS) != 0) signal[t] = 0.0;
				else if(exposure.sampleFlag[bol] != 0) signal[t] = 0.0;
				else signal[t] = frame[t].data[bol];
			}
			Arrays.fill(signal, frames, signal.length, 0.0);
			ArrayUtil.uncheckedFFT(signal, spectrum);

			//for(int f=0; f<spectrum.length; f++) spectrum[f].scale(Math.exp(-A*f*f));
			int fmin = Math.min(spectrum.length, (int) Math.ceil(2.35*sigma_f));
			for(int f=fmin; f<spectrum.length; f++) spectrum[f].scale(0.0);

			ArrayUtil.uncheckedBackFFT(spectrum, signal);

			for(int t=0; t<frames; t++) if(frame[t] != null) frame[t].data[bol] -= (float) signal[t];
		}
	}	
	
	public void downsample(int n) {
		System.err.print(" Downsampling by " + n);

		int windowSize = (int)Math.round(1.82*ArrayUtil.getEquivalentWindowWidth("Hann")*n);
		double[] w = ArrayUtil.getWindowFunction("Hann", windowSize);

		LabocaFrame[] newFrame = new LabocaFrame[(frame.length-windowSize)/n+1]; 
		double[] newValue = new double[channels];

		System.err.println(" to " + newFrame.length + " frames.");


		for(int T=0, fromt=0; fromt+windowSize <= frames; T++, fromt += n) {
			Arrays.fill(newValue, 0.0);
			boolean hasEmpty = false;
			for(int t=fromt, blockt=0; blockt<windowSize; t++, blockt++) 
				for(int c=0; c<channels; c++) {
					if(frame[t] == null) { hasEmpty = true; break; }
					else newValue[c] += w[blockt] * frame[t].data[c];
				}
			newFrame[T] = hasEmpty ? null : frame[fromt+windowSize/2];
			if(!hasEmpty) for(int c=0; c<channels; c++) newFrame[T].data[c] = (float) newValue[c];
		}

		integrationTime *= n;
		frame = newFrame;
		frames = frame.length;
	}

	public void checkForNaNs() {
		System.err.print("?");
		for(int t=0; t<frame.length; t++) for(int c=0; c<channel.length; c++) {
			/*
	    if(Float.isNaN(frame[t].data[c])) {
		System.err.println("NaN: " + t + "," + c);
		System.exit(1);

	    if(Float.isInfinite(frame[t].data[c])) {
		System.err.println("Inf: " + t + "," + c);
		System.exit(1);
	    }
			 */
			if(frame[t] != null) if(frame[t].data[c] > 10.0) {
				System.err.println("Big: " + t + "," + c);
				System.exit(1);
			}
		}

	}

	public void shift(double dt) {
		double[][] coords = new double[7][frames];
		for(int t=0; t<frames; t++) {
			if(frame[t] != null) {
				if((frame[t].flag & LabocaFrame.BAD_DATA) == 0) {
					coords[0][t] = frame[t].RA;
					coords[1][t] = frame[t].DEC;
					coords[2][t] = frame[t].LST;
					coords[3][t] = frame[t].AZ;
					coords[4][t] = frame[t].EL;
					coords[5][t] = frame[t].dX;
					coords[6][t] = frame[t].dY;
				}
				else for(int i=0; i<7; i++) coords[i][t] = Double.NaN;
			}
			else for(int i=0; i<7; i++) coords[i][t] = Double.NaN;
		}

		for(int i=0; i<7; i++) coords[i] = ArrayExtras.shift(coords[i], dt/integrationTime, 0.5);

		for(int t=0; t<frames; t++) if(frame[t] != null) {
			if(Double.isNaN(coords[0][t])) frame[t] = null;
			else {
				frame[t].RA = coords[0][t];
				frame[t].DEC = coords[1][t];
				frame[t].LST = coords[2][t];
				frame[t].AZ = coords[3][t];
				frame[t].EL = coords[4][t];
				frame[t].dX = coords[5][t];
				frame[t].dY = coords[6][t];		
			}
		}
	}


	public void printASCIITimeStream() throws IOException {
		PrintStream out = new PrintStream(new BufferedOutputStream(new FileOutputStream(MiniCRUSH.workPath + File.separator + serialNo + "-" + subscanNo + ".tms"), 1000000));
		out.println(Util.e3.format(1.0/integrationTime));
		for(int t=0; t<frames; t++) {
			boolean isEmpty = true;
			if(frame[t] != null) if((frame[t].flag & LabocaFrame.BAD_DATA) == 0) {
				isEmpty = false;
				for(int c=0; c<channels; c++) 
					out.print(Util.e5.format((frame[t].sampleFlag[c] & LabocaFrame.SAMPLE_SPIKE_FLAGS) != 0 ? 0.0 : frame[t].data[c]) + "\t");
			}
			if(isEmpty) for(int c=0; c<channels; c++) out.print(Util.e5.format(0.0) + "\t");
			out.println();
				
		}
		out.flush();
		out.close();
	}
	
	public double[][] getCovariance() {
		System.err.print(" Calculating Covariance Matrix (this may take a while...)");
		
		double[][] covar = new double[channels][channels];
		int[][] n = new int[channels][channels];
		
		for(int t=0; t<frames; t++) if(frame[t] != null) if((frame[t].flag & LabocaFrame.SOURCE_FLAGS) == 0) {
			LabocaFrame exposure = frame[t];
			float[] data = exposure.data;

			for(int c1=0; c1<channels; c1++) if(channel[c1].flag == 0) if(exposure.sampleFlag[c1] == 0)
				for(int c2=c1+1; c2<channels; c2++) if(channel[c2].flag == 0) if(exposure.sampleFlag[c2] == 0) {
					covar[c1][c2] += data[c1] * data[c2];
					n[c1][c2]++;
				}
		}
			
		for(int c1=0; c1<channels; c1++) for(int c2=c1+1; c2<channels; c2++) {
			covar[c1][c2] *= Math.sqrt(channel[c1].weight * channel[c2].weight) / n[c1][c2];
			covar[c2][c1] = covar[c1][c2];
		}

		System.err.println();
		
		return covar;
	}
    	
	public static double[][] getGroupCovariance(int[][] channelSelection, double[][] covar) {
		if(channelSelection == null) return null;

		int n = 0;
		for(int g=0; g<channelSelection.length; g++) n+= channelSelection[g].length;

		double[][] groupedCovar = new double[n][n];

		for(int k=0; k<n; k++) Arrays.fill(groupedCovar[k], Double.NaN);

		for(int g1=0, k1=0; g1<channelSelection.length; g1++) {
			for(int c1=0; c1<channelSelection[g1].length; c1++,k1++)
				for(int g2=0, k2=0; g2<channelSelection.length; g2++) { 
					for(int c2=0; c2<channelSelection[g2].length; c2++,k2++) {
						groupedCovar[k1][k2] = covar[channelSelection[g1][c1]][channelSelection[g2][c2]];
					}
				}
		}

		return groupedCovar;
	}
	
	public double[][] getFullCovariance(double[][] covar) {
		double[][] fullCovar = new double[320][320];
		
		for(int i=0; i<fullCovar.length; i++) Arrays.fill(fullCovar[i], Double.NaN);
		
		for(int c1=0; c1<channels; c1++) for(int c2=0; c2<channels; c2++)
			fullCovar[channel[c1].backendIndex][channel[c2].backendIndex] = covar[c1][c2];
			
		return fullCovar;
	}
	
	public void writeCovariance(String name, double[][] covar) throws IOException, FitsException {
		if(!name.endsWith(".fits")) name += "-" + serialNo + ".fits";
		if(covar == null) return;
		Fits fits = new Fits();
		BasicHDU hdu = Fits.makeHDU(covar);
		fits.addHDU(hdu);
		fits.write(new BufferedDataOutputStream(new FileOutputStream(name)));	
		System.err.println(" Written " + name);
	}
	
	
	float[] arrayCorrelated, arrayGain;
	float[][] cableCorr, cableGain;
	float[][] ampCorr, ampGain;
	float[][] boxCorr, boxGain;
	float[][] waferCorr, waferGain;
	float[][] regionCorr, regionGain;
	float[][] accelTerm, accelCoeff;
	float[] allXGain, allYGain, skyX, skyY;

	float[][] createCorrelatedStorage(String channelGroupSpec) {
		int[][] channelGroup = getChannelGroup(channelGroupSpec);
		if(channelGroup == null) return null;
		 
		float[][] groupCorr = new float[channelGroup.length][];
		if(channelGroup != null) {
			groupCorr = new float[channelGroup.length][];
				for(int g=0; g<channelGroup.length; g++) groupCorr[g] = new float[frames];
		}
		return groupCorr;
	}
	
	float[][] createGainStorage(String channelGroupSpec) {
		int[][] channelGroup = getChannelGroup(channelGroupSpec);
		if(channelGroup == null) return null;
		
		float[][] groupGain = new float[channelGroup.length][];
		for(int g=0; g<channelGroup.length; g++) {
			groupGain[g] = new float[channelGroup[g].length];	
			Arrays.fill(groupGain[g], 1.0F);
		}

		return groupGain;	
	}
	
	public void createStorage() {	
		cableCorr = createCorrelatedStorage("cables");
		cableGain = createGainStorage("cables");

		ampCorr = createCorrelatedStorage("amps");
		ampGain = createGainStorage("amps");

		boxCorr = createCorrelatedStorage("boxes");
		boxGain = createGainStorage("boxes");
		
		waferCorr = createCorrelatedStorage("wafers");
		waferGain = createGainStorage("wafers");
		
		regionCorr = createCorrelatedStorage("regions");
		regionGain = createGainStorage("regions");
		
		accelTerm = new float[2][frames];
		accelCoeff = new float[2][allChannels.length];

		allXGain = new float[allChannels.length];
		allYGain = new float[allChannels.length];

		arrayCorrelated = new float[frames];
		arrayGain = new float[allChannels.length];
		
		skyX = new float[frames];
		skyY = new float[frames];
				
		Vector2D[] accel = getAccelerations();

		for(int t=0; t<frames; t++) {
			accelTerm[0][t] = accel[t] == null ? 0.0F : (float)Math.abs(accel[t].x);
			accelTerm[1][t] = accel[t] == null ? 0.0F : (float)Math.abs(accel[t].y);
		}

		Arrays.fill(accelCoeff[0], 1.0F);
		Arrays.fill(accelCoeff[1], 1.0F);

		Vector2D arraycenter = channel[referenceChannel].position;

		for(int s=0; s<allChannels.length; s++) {
			LabocaPixel pixel = channel[allChannels[s]];
			Vector2D position = pixel.position;
			allXGain[s] = (float)(position.x - arraycenter.x);
			allYGain[s] = (float)(position.y - arraycenter.y);
			arrayGain[s] = (float) pixel.gain;
		}
	}
	
	
	public void getArrayCorrelated(boolean flag, boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[C]" : "C";
		getCorrelated(allChannels, arrayGain, arrayCorrelated, robust);
		if(solveGains) {
			getGains(allChannels, arrayGain, arrayCorrelated, 0.1, 10.0, flag ? LabocaPixel.FLAG_GAIN : 0, false, LabocaPixel.GAINS_BIDIRECTIONAL, gainRobust);
			applyGains(allChannels, arrayGain);
			if(flag) {
				census();
				comments += goodChannelCount + " ";
			}
		}
	}
	
	
	
	

	
	
	public void getCorrs(int[][] channelGroup, float[][] gCorr, float[][] gGain, boolean robust) {		
		for(int g=0; g<channelGroup.length; g++) getCorrelated(channelGroup[g], gGain[g], gCorr[g], robust);		
	}
	
	

	
	public void getGains(int[][] channelGroup, float[][] gCorr, float[][] gGain, double loGain, double highGain, int flagPattern, boolean robust) {		
		for(int g=0; g<channelGroup.length; g++) getGains(channelGroup[g], gGain[g], gCorr[g], loGain, highGain, flagPattern, robust, LabocaPixel.GAINS_BIDIRECTIONAL);
		census();
		comments += goodChannelCount + " ";	
	}
	
	

	
	
	


	public void getDrifts(int i, int k, int nFrames, boolean robust) {
		comments += robust ? "[D]" : "D";

		int p = (int)Math.round(Math.log(nFrames) / Math.log(2.0));
		if(p < 1) p = 1;
		driftN = (int)Math.round(Math.pow(2.0, p));
		
		comments += "[" + driftN + "]";
		
		if(!robust && map != null) {
			double crossingtime = map.beamFWHM / aveScanSpeed;
			for(int c=0; c<channels; c++) channel[c].filtering *= 1.0 - crossingtime / (driftN * integrationTime); 
		}
			
		for(int t=0; t < frames;) {
			int tot = t+driftN;
			if(frames < tot+driftN) tot = frames;
			removeOffsets(allChannels, t, tot, robust);
			t = tot;
		}
	}
	
	public void resetFiltering() {
		// Make the reset channel filtering the highpass-filtering value...
		for(int c=0; c<channels; c++) channel[c].filtering = 1.0;
	}
	
	public void calcGroupFiltering(int[] channelGroup, float[] gain) {
		int n = 0;
		double sumw = 0.0;
		for(int k=0; k<channelGroup.length; k++) {
			LabocaPixel pixel = channel[channelGroup[k]];
			if((pixel.flag == 0)) { 
				sumw += pixel.weight * gain[k] * gain[k]; 
				if(pixel.weight > 0.0) n++; 
			}		
		}
		
		if(n > 1) for(int k=0; k<channelGroup.length; k++) {
			LabocaPixel pixel = channel[channelGroup[k]];
			if((pixel.flag == 0)) pixel.filtering *= 1.0 - pixel.weight * gain[k] * gain[k] / sumw;			
		}
	}
	
	
	private double averageFiltering;
	public void calcAverageFiltering() {
		double sum = 0.0, sumw = 0.0;
		for(int c=0; c<channels; c++) if(channel[c].flag == 0) {
			double G = channel[c].gain * channel[c].mainBeamEfficiency;
			double w = channel[c].weight * G * G;
			sum += w * channel[c].filtering;
			sumw += w;			
		}
		if(sumw > 0.0) averageFiltering = sum / sumw;
		else averageFiltering = 0.0;
	}
	


	
	
	public void getTwisting(boolean robust) {
		float[][] twist = new float[cableChannels.length][frames];
		float[][] twistGain = new float[cableChannels.length][];
		
		for(int g=0; g<twistGain.length; g++) {
			twistGain[g] = new float[cableChannels[g].length];
			
			double avepin = 0.0, sumw = 0.0;
			for(int c=0; c<twistGain[g].length; c++) {
				double G = cableGain[g][c];
				LabocaPixel pixel = channel[cableChannels[g][c]];
				if(pixel.flag == 0) {
					avepin += pixel.weight * G * pixel.pin;
					sumw += pixel.weight * G;
				}
			}
			if(sumw > 0.0) avepin /= sumw;
			
			for(int c=0; c<twistGain[g].length; c++)
				twistGain[g][c] = cableGain[g][c] * (channel[cableChannels[g][c]].pin - (float)avepin);
		}
		
		comments += robust ? "[t]" : "t";
		for(int g=0; g<cableChannels.length; g++) getCorrelated(cableChannels[g], twistGain[g], twist[g], robust);
	}
	
	
	public void getBoxCorrelated(boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[B]" : "B";
		getCorrs(boxChannels, boxCorr, boxGain, robust);
		if(solveGains) {
			getGains(boxChannels, boxCorr, boxGain, 0.0, 10.0, LabocaPixel.FLAG_BOX, gainRobust);
			for(int g=0; g<boxChannels.length; g++) 
				for(int k=0; k<boxChannels[g].length; k++) 
					channel[boxChannels[g][k]].boxGain = boxGain[g][k];			
		}	
		
		/*
		System.err.println();
		for(int i=0; i<boxChannels.length; i++) {
			System.err.print("Box " + (i+1) + ": ");
			for(int k=0; k<boxChannels[i].length; k++) {
				LabocaPixel pixel = channel[boxChannels[i][k]];
				System.err.print((pixel.backendIndex+1) + " ");
			}
			System.err.println();
		}
		*/
	}
	
	public void getCableCorrelated(boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[c]" : "c";
		getCorrs(cableChannels, cableCorr, cableGain, robust);
		if(solveGains) {
			getGains(cableChannels, cableCorr, cableGain, 0.0, 10.0, LabocaPixel.FLAG_CABLE, gainRobust);
			for(int g=0; g<cableChannels.length; g++) 
				for(int k=0; k<cableChannels[g].length; k++) 
					channel[cableChannels[g][k]].cableGain = cableGain[g][k];
		}
		
		/*
		System.err.println();
		for(int i=0; i<cableChannels.length; i++) {
			System.err.print("Cable " + (i+1) + ": ");
			for(int k=0; k<cableChannels[i].length; k++) {
				LabocaPixel pixel = channel[cableChannels[i][k]];
				System.err.print((pixel.backendIndex+1) + " ");
			}
			System.err.println();
		}
		*/
		
	}

	
	public void getAmpCorrelated(boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[a]" : "a";
		getCorrs(ampChannels, ampCorr, ampGain, robust);
		if(solveGains) {
			getGains(ampChannels, ampCorr, ampGain, 0.0, 10.0, LabocaPixel.FLAG_AMP, gainRobust);	
			for(int g=0; g<ampChannels.length; g++) 
				for(int k=0; k<ampChannels[g].length; k++) 
					channel[ampChannels[g][k]].ampGain = ampGain[g][k];
		}
	}

	public void getWaferCorrelated(boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[Q]" : "Q";
		getCorrs(waferChannels, waferCorr, waferGain, robust);
		if(solveGains) {
			getGains(waferChannels, waferCorr, waferGain, 0.0, 10.0, LabocaPixel.FLAG_WAFER, gainRobust);
			for(int g=0; g<waferChannels.length; g++) 
				for(int k=0; k<waferChannels[g].length; k++) 
					channel[waferChannels[g][k]].waferGain = waferGain[g][k];
		}
	}
	
	public void getRegionCorrelated(boolean solveGains, boolean robust, boolean gainRobust) {
		comments += robust ? "[r]" : "r";
		getCorrs(regionChannels, regionCorr, regionGain, robust);
		if(solveGains) {
			getGains(regionChannels, regionCorr, regionGain, 0.0, 10.0, LabocaPixel.FLAG_REGION, gainRobust);
			for(int g=0; g<regionChannels.length; g++) 
				for(int k=0; k<regionChannels[g].length; k++) 
					channel[regionChannels[g][k]].regionGain = regionGain[g][k];
		}
	}
	
	
	
	// sum wg (x-x0) C / sum w g^2 (x-x0)^2 = 0
	// sum wgx = x0 sum wg
	// x0 = sum wgx / sum wg
	
	public void getSkyGradients(boolean robust) {
		float[] xGain = new float[allXGain.length];
		float[] yGain = new float[allYGain.length];

		Vector2D center = new Vector2D();
		double sumw = 0.0;
		
		for(int c=0; c<channels; c++) {
			LabocaPixel pixel = channel[c];
			if(pixel.flag == 0) {
				center.addMultipleOf(pixel.position, pixel.weight * pixel.gain);
				sumw += pixel.weight * pixel.gain;
			}
		}
		if(sumw > 0.0) center.scale(1.0 / sumw);
		
		for(int c=0; c<channels; c++) {
			LabocaPixel pixel = channel[c];
			xGain[c] = (float) (pixel.gain * (pixel.position.x - center.x));
			yGain[c] = (float) (pixel.gain * (pixel.position.y - center.y));
		}
		
		comments += robust ? "[S]" : "S";
		getCorrelated(allChannels, xGain, skyX, robust);
		getCorrelated(allChannels, yGain, skyY, robust);
	}
	
	public void getAccelCoeffs(boolean robust) {
		comments += robust ? "[a]" : "a";
		getGains(allChannels, accelCoeff[0], accelTerm[0], 0.1, 10.0, LabocaPixel.FLAG_XACCEL, robust, LabocaPixel.GAINS_BIDIRECTIONAL);
		getGains(allChannels, accelCoeff[1], accelTerm[1], 0.1, 10.0, LabocaPixel.FLAG_YACCEL, robust, LabocaPixel.GAINS_BIDIRECTIONAL);
		census();
	}
	
	
	public double correlatedFiltering = Double.NaN;
	public double skyFiltering = Double.NaN;
	public double boxesFiltering = Double.NaN;
	public double cablesFiltering = Double.NaN;
	public double twistingFiltering = Double.NaN;
	public double ampFiltering = Double.NaN;
	public double waferFiltering = Double.NaN;
	
	
	public double getPointSourceFiltering(int[][] channelGroup) {
		double phi = 0.0;
		double allsumw = 0.0;
		
		for(int g=0; g<channelGroup.length; g++) {
			double sumw=0.0;
			int n=0;
			for(int k=0; k<channelGroup[k].length; k++) {
				LabocaPixel pixel = channel[channelGroup[g][k]];
				if(pixel.flag == 0) {
					n++;
					sumw += pixel.weight;
				}	
			}
			if(n > 0) phi += sumw / n;
			allsumw += sumw;
		}
		
		if(allsumw > 0.0) phi /= allsumw;
		
		return phi;		
	}
	
	
	private static TauInterpolator tauInterpolator = new TauInterpolator();
	static {
		System.err.println("Initializing tau interpolator.");
		try { tauInterpolator.readData("laboca/laboca-taus-master.dat"); }
		catch(IOException e) { System.err.println("WARNING! " + e.getMessage()); }		
	}
	
	static String path = "/home/pumukli/data/laboca";
	// filtering will apply to points below clipping and blanking levels when models are solved before source...
}

